'''
# Rearrangement plotting

'''

import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import seaborn
import random
import scipy

import wgs_analysis.plots as plots
import wgs_analysis.refgenome as refgenome
import wgs_analysis.plots.colors


default_rearrangement_types = ['foldback', 'deletion', 'duplication', 'inversion', 'balanced', 'unbalanced', 'complex']
default_rearrangement_type_colors = seaborn.color_palette('Dark2', len(default_rearrangement_types))


def chromosome_type_plot(ax, breakends, bin_size=20000000, rearrangement_types=None, full_genome=True):
    if rearrangement_types is None:
        rearrangement_types = default_rearrangement_types
        rearrangement_type_colors = default_rearrangement_type_colors
    else:
        rearrangement_type_colors = seaborn.color_palette('Dark2', len(rearrangement_types))

    breakends = breakends.loc[(breakends['chromosome'].isin(refgenome.info.chromosomes))].copy()

    breakends.set_index('chromosome', inplace=True)
    breakends['chromosome_start'] = refgenome.info.chromosome_start
    breakends.reset_index(inplace=True)

    breakends['plot_position'] = breakends['position'] + breakends['chromosome_start']
    breakends['plot_bin'] = breakends['plot_position'] / bin_size
    breakends['plot_bin'] = breakends['plot_bin'].astype(int)

    if full_genome:
        plot_bin_starts = np.arange(0, refgenome.info.chromosome_end.max(), bin_size)
    else:
        min_x, max_x = breakends['plot_position'].min(), breakends['plot_position'].max()
        plot_bin_starts = np.arange(min_x, max_x, bin_size)
    plot_bins = (plot_bin_starts / bin_size).astype(int)

    breakends = breakends[['prediction_id', 'plot_bin', 'rearrangement_type']].drop_duplicates()

    count_table = breakends.groupby(['plot_bin', 'rearrangement_type']).size().unstack()
    count_table = count_table.reindex(plot_bins).fillna(0)

    accum_counts = None
    for idx, rearrangement_type in enumerate(rearrangement_types):
        if rearrangement_type not in count_table.columns:
            continue
        ax.bar(plot_bin_starts, count_table[rearrangement_type].values,
               bottom=accum_counts, width=bin_size,
               facecolor=rearrangement_type_colors[idx],
               edgecolor=rearrangement_type_colors[idx])
        if accum_counts is None:
            accum_counts = count_table[rearrangement_type].values
        else:
            accum_counts += count_table[rearrangement_type].values

    if full_genome:
        ax.set_xlim((-0.5, refgenome.info.chromosome_end.max()))
    else:
        ax.set_xlim((min_x, max_x))
    ax.set_xlabel('chromosome')
    ax.set_xticks([0] + list(refgenome.info.chromosome_end.values))
    ax.set_xticklabels([])
    ax.set_ylabel('count')
    ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(refgenome.info.chromosome_mid))
    ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(refgenome.info.chromosomes))
    ax.xaxis.grid(False, which="minor")


def chromosome_type_plot_legend(ax, rearrangement_types=None):
    if rearrangement_types is None:
        rearrangement_types = default_rearrangement_types
        rearrangement_type_colors = default_rearrangement_type_colors
    else:
        rearrangement_type_colors = seaborn.color_palette('Dark2', len(rearrangement_types))

    ax.legend([plt.Circle((0, 0), color=c) for c in rearrangement_type_colors],
              list(rearrangement_types), loc="upper right")


def type_size_plot(data, type_col='type'):
    data = data.copy()
    data['type'] = data['type'].astype('category')
    data['rearrangement_type'] = pd.Categorical(
        data['rearrangement_type'],
        categories=[
            'duplication',
            'inversion',
            'balanced',
            'foldback',
            'deletion',
            'unbalanced',
        ])
    data['type'] = pd.Categorical(
        data['type'],
        categories=[
            'duplication',
            'inversion',
            'deletion',
            'translocation',
        ])

    seaborn.catplot(
        x=type_col, hue='size_class', kind='count',
        data=data, sharex=False,
        aspect=3, height=2.5)
    seaborn.despine(trim=True)


def breakpoint_adjacent_density_plot(ax, breakends):
    breakends = breakends.drop_duplicates(['prediction_id', 'prediction_side'])

    breakends = breakends.loc[(breakends['chromosome'].isin(refgenome.info.chromosomes))]

    breakends.set_index('chromosome', inplace=True)
    breakends['chromosome_start'] = refgenome.info.chromosome_start
    breakends['chromosome_color'] = pd.Series(plots.colors.create_chromosome_color_map())
    breakends.reset_index(inplace=True)

    breakends['plot_position'] = breakends['position'] + breakends['chromosome_start']

    assert not breakends['chromosome_color'].isnull().any()

    ax.scatter(
        breakends['plot_position'], breakends['adjacent_density'],
        facecolors=list(breakends['chromosome_color']),
        edgecolors=list(breakends['chromosome_color']),
        alpha=1.0, s=10, lw=0
    )

    ax.set_xlim((-0.5, refgenome.info.chromosome_end.max()))
    ax.set_xlabel('chromosome')
    ax.set_xticks([0] + list(refgenome.info.chromosome_end.values))
    ax.set_xticklabels([])
    ax.set_ylabel('count')
    ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(refgenome.info.chromosome_mid))
    ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(refgenome.info.chromosomes))
    ax.xaxis.grid(False, which="minor")

    ax.set_ylim(-.05 * breakends['adjacent_density'].max(), 1.1 * breakends['adjacent_density'].max())
    ax.set_ylabel('breakpoint density')


def plot_polar_arc(ax, u1, u2, r, rt, color='k', lw=0.5):
    '''
    Plot an arc in polar coordinates
    '''

    if u1 == u2:
        u2 += np.finfo(float).eps

    theta_1 = u1 * 2. * np.pi
    theta_2 = u2 * 2. * np.pi

    x1 = r * np.cos(theta_1)
    y1 = r * np.sin(theta_1)

    x2 = r * np.cos(theta_2)
    y2 = r * np.sin(theta_2)

    xt = rt * np.cos(0.5 * (theta_1 + theta_2))
    yt = rt * np.sin(0.5 * (theta_1 + theta_2))

    m1 = (y1 - yt) / (x1 - xt)
    m2 = (y2 - yt) / (x2 - xt)

    a = np.array([
                  [0, 0, 1, 0, 0, 0],
                  [0, 0, 0, 0, 0, 1],
                  [1, 1, 1, 0, 0, 0],
                  [0, 0, 0, 1, 1, 1],
                  [0, m1, 0, 0, -1, 0],
                  [2*m2, m2, 0, -2, -1, 0]
                 ])

    b = np.array([x1, y1, x2, y2, 0, 0])

    try:
        c = np.linalg.solve(a, b)
    except:
        print(u1, u2, r, rt)
        raise

    xs = list()
    ys = list()

    for t in np.linspace(0, 1, 101):
        basis = np.array([t**2, t, 1])
        x = np.dot(c[:3], basis)
        y = np.dot(c[3:], basis)
        xs.append(x)
        ys.append(y)

    xs = np.array(xs)
    ys = np.array(ys)
    
    r = np.sqrt(xs**2 + ys**2)
    theta = np.arctan(ys / xs)
    theta[xs < 0] = theta[xs < 0] + np.pi

    ax.plot(theta, r, color=color, lw=lw)


def breakpoint_circos_plot(ax, breakpoints):
    genome_length = refgenome.info.chromosome_lengths.sum()

    for breakend in ('1', '2'):
        breakpoints.set_index('chromosome_' + breakend, inplace=True)
        breakpoints['chromosome_' + breakend + '_start'] = refgenome.info.chromosome_start
        breakpoints.reset_index(inplace=True)

        breakpoints['plot_position_' + breakend] = breakpoints['position_' + breakend] + breakpoints['chromosome_' + breakend + '_start']
        breakpoints['plot_fraction_' + breakend] = breakpoints['plot_position_' + breakend] / genome_length

    for idx in breakpoints.index:
        plot_polar_arc(ax, breakpoints.loc[idx, 'plot_fraction_1'], breakpoints.loc[idx, 'plot_fraction_2'], 1., 0.2)

    chrom_colors = plots.colors.create_chromosome_color_map()
    for chrom, start, end in zip(refgenome.info.chromosomes, refgenome.info.chromosome_start, refgenome.info.chromosome_end):
        start = 2. * np.pi * start / genome_length
        end = 2. * np.pi * end / genome_length
        ax.plot(np.linspace(start, end, 1001), [1.] * 1001, lw=20, color=chrom_colors[chrom], solid_capstyle='butt')

    xticks = 2. * np.pi * np.array([0] + list(refgenome.info.chromosome_end.values)) / genome_length
    tickmids = 2. * np.pi * refgenome.info.chromosome_mid / genome_length

    ax.set_xlabel('chromosome')
    ax.set_xticks(xticks)
    ax.set_xticklabels([])
    ax.set_yticks([])
    ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(tickmids))
    ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(refgenome.info.chromosomes))
    ax.xaxis.grid(False)

    ax.set_ylim(0., 1.)
    ax.set_ylabel('')


def plot_breakends(ax, breakends, chromosome=None, lw=0.5):
    """ Plot breakpoint flags and arcs on a genome axis

    Args:
        ax: matplotlib axis
        breakends: breakend table
    
    Kwargs:
        lw: line width for arcs

    The breakends table should have the following columns:
        - prediction_id
        - chromosome
        - position
        - strand

    """
    chromosome_info = refgenome.info.chromosome_info[['chr', 'chromosome_start']].copy()
    chromosome_info['chromosome'] = chromosome_info['chr']

    if chromosome is None:
        plot_data = breakends.merge(chromosome_info)
        plot_data = plot_data[plot_data['chr'].isin(refgenome.info.chromosomes)]
        plot_data['plot_position'] = plot_data['position'] + plot_data['chromosome_start']

    else:
        plot_data = breakends.copy()
        plot_data['plot_position'] = plot_data['position']

    xlim = ax.get_xlim()
    xrng = xlim[1] - xlim[0]

    ylim = ax.get_ylim()
    yrng = ylim[1] - ylim[0]

    arrow_length = 0.01 * xrng

    head_width = yrng / 30.
    head_length = xrng / 100.

    prediction_ids = list(plot_data['prediction_id'].unique())
    random.shuffle(prediction_ids)
    color_palette = seaborn.color_palette('hls', len(prediction_ids))
    prediction_colors = dict(zip(prediction_ids, color_palette))

    for idx in plot_data.index:
        if chromosome is not None and plot_data.loc[idx, 'chromosome'] != chromosome:
            continue
        prediction_id = plot_data.loc[idx, 'prediction_id']
        x = plot_data.loc[idx, 'plot_position']
        strand = plot_data.loc[idx, 'strand']
        y = np.random.uniform(ylim[0] + 0.75*yrng, ylim[0] + 0.95*yrng)
        offset = (arrow_length, -arrow_length)[strand == '+']
        color = prediction_colors[prediction_id]
        ax.arrow(x, y, offset, 0, color=color, lw=lw, alpha=1.0, head_width=head_width, head_length=head_length)
        ax.plot([x, x], [ylim[0], ylim[1]], color=color, lw=lw, ls='-', alpha=1.0)

    for prediction_id, pred_breakends in plot_data.groupby('prediction_id'):
        if len(pred_breakends.index) != 2:
            continue
        if chromosome is not None and set(pred_breakends['chromosome'].values) != set([chromosome]):
            continue
        pos1 = pred_breakends['plot_position'].min()
        pos2 = pred_breakends['plot_position'].max()
        posmid = (pos1 + pos2) / 2.
        height = 0.5 * yrng * (pos2 - pos1) / xrng
        ypos = ylim[1]
        yposmid = ypos + height
        spl = scipy.interpolate.make_interp_spline([pos1, posmid, pos2], [ypos, yposmid, ypos], k=2)
        pos = np.linspace(pos1, pos2, 100)
        color = prediction_colors[prediction_id]
        ax.plot(pos, spl(pos), ls='-', lw=lw, color=color)

    ax.set_ylim((ylim[0], ylim[1] + 0.5 * (ylim[1] - ylim[0])))


def plot_arrow(ax, x0, y0, xlength, ylength, color, 
               arrow_width = 0.05, arrow_edgecolor = (0,0,0,0)):
    """ Plot an arrow in `ax` from (x0, y0) with lengths of (xlength, ylength)
    """
    ax.arrow(x0, y0, xlength, ylength, facecolor=color, 
             width=arrow_width, head_width=arrow_width, 
             edgecolor=arrow_edgecolor, length_includes_head=True) 

def is_pos_in_xlim(chrom, pos, plot_chrom, xlim):
    """ Check if given chrom and pos are in plot limits
    """
    if chrom != plot_chrom:
        return False
    xmin, xmax = xlim
    if pos < xmin: return False
    if pos > xmax: return False
    return True


def plot_sv_on_ax(ax, sv, chromosome):
    """ Plot adjacencies in a given Axes instance

    Args:
        ax: matplotlib axis
        sv: consensus csv-like sv table
    
    Kwargs:
        chromosome: chromosome that `ax` is based on

    The breakends table should have the following columns:
        - chromosome_[12]
        - position_[12]
        - strand_[12]
        - gene_name_[12]
        - gene_location_[12]
        - type
    """
    svs = sv.copy()
    svcolors = {
        'deletion':'blue', 'insertion':'green', 'inversion':'purple', 'duplication':'red', 'translocation':'black'
    }
    linewidth = 1
    codes = [
        matplotlib.path.Path.MOVETO,
        matplotlib.path.Path.CURVE4,
        matplotlib.path.Path.CURVE4,
        matplotlib.path.Path.CURVE4,
    ]
    svs['chromosome_1'] = svs['chromosome_1'].str.replace('chr', '')
    svs['chromosome_2'] = svs['chromosome_2'].str.replace('chr', '')
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    x_offset = (xmax - xmin) / 100
    chromosomes = [str(c) for c in range(1, 22+1)] + ['X', 'Y']

    for i, (rix, row) in enumerate(svs.iterrows()):
        # sv features
        chrom1, chrom2 = row['chromosome_1'], row['chromosome_2']
        assert chrom1 in chromosomes and chrom2 in chromosomes
        pos1, pos2 = row['position_1'], row['position_2']
        strand1, strand2 = row['strand_1'], row['strand_2']
        gene1, gene2 = row['gene_name_1'], row['gene_name_2']
        geneloc1, geneloc2 = row['gene_location_1'], row['gene_location_2']
        
        if chrom1 == chrom2 and pos1 > pos2:
            pos1, pos2, strand1, strand2 = pos2, pos1, strand2, strand1
            gene1, gene2, geneloc1, geneloc2 = gene2, gene1, geneloc2, geneloc1
        elif chromosomes.index(chrom2) < chromosomes.index(chrom1):
            chrom1, chrom2, pos1, pos2, strand1, strand2 = chrom2, chrom1, pos2, pos1, strand2, strand1
            gene1, gene2, geneloc1, geneloc2 = gene2, gene1, geneloc2, geneloc1

        svtype = row['type']
        if (not is_pos_in_xlim(chrom1, pos1, chromosome, ax.get_xlim()) and 
            not is_pos_in_xlim(chrom2, pos2, chromosome, ax.get_xlim())): 
            continue
        svcolor = svcolors[svtype]
        arrow_facecolor = svcolor
        length = abs(pos2 - pos1)
        if 'length' in row:
            length = row['length']
        svtext = f'{length}bp'

        x1, x2 = pos1, pos2
        x1_between_xlim = chrom1 == chromosome and x1 >= xmin and x1 < xmax
        x2_between_xlim = chrom2 == chromosome and x2 >= xmin and x2 < xmax
        y_diff = ymax - ymin
        y_top = ymin + 0.6 * y_diff # ymin + 0.3 * y_diff
        offset_base = int(y_diff * 0.15)
        offset_with_ix = offset_base * 0.7 * (i+1)
        y_offset = max(offset_base, offset_with_ix)
        arrow_width = y_diff / 100

        # plot sv
        x1_component = [(xmin, y_top+y_offset), (xmin+1, y_top+y_offset)]
        x2_component = [(xmax-1, y_top+y_offset), (xmax, y_top+y_offset)]
        if x1_between_xlim:
            if svtype == 'TRA':
                # svtext = f'~chr{chrom2}:{pos2}'
                svtext = f'~ {gene2}: {geneloc2}'
                
            x1_component = [(x1, y_top), (x1, y_top+y_offset)]
            ax.plot([x1, x1], [0, y_top], color=svcolor, linewidth=linewidth)
            if strand1 == '+': plot_arrow(ax, x1-x_offset, y_top, +x_offset, 0, color=arrow_facecolor, arrow_width=arrow_width)
            else:              plot_arrow(ax, x1+x_offset, y_top, -x_offset, 0, color=arrow_facecolor, arrow_width=arrow_width)
        else: 
            pass 

        if x2_between_xlim:
            if svtype == 'TRA':
                # svtext = f'~chr{chrom1}:{pos1}'
                svtext = f'~ {gene1}: {geneloc1}'
            x2_component = [(x2, y_top+y_offset), (x2, y_top)]
            ax.plot([x2, x2], [0, y_top], color=svcolor, linewidth=linewidth)            
            if strand2 == '+': plot_arrow(ax, x2-x_offset, y_top, +x_offset, 0, color=arrow_facecolor, arrow_width=arrow_width)
            else:              plot_arrow(ax, x2+x_offset, y_top, -x_offset, 0, color=arrow_facecolor, arrow_width=arrow_width)
        else: 
            pass

        verts = x1_component + x2_component
        path = matplotlib.path.Path(verts, codes)
        patch = matplotlib.patches.PathPatch(path, facecolor='none', lw=0.5, edgecolor=svcolor)
        ax.add_patch(patch)
        y_text = (y_top + y_offset)# + 0.15 * y_diff

        if x1_between_xlim and x2_between_xlim:
            x_text = (x1 + x2) // 2
        elif x1_between_xlim and not x2_between_xlim:
            x_text = (x1 + xmax) // 2
        elif not x1_between_xlim and x2_between_xlim:
            x_text = (xmin + x2) // 2
        else:
            raise ValueError('coordinates not between x1 or x2')
        ax.text(x_text, y_text, svtext, color=svcolor, ha='center', fontdict={'size':11})
        if ymax < y_text: 
            ax.set_ylim((ymin, y_text))
            ax.spines['left'].set_bounds(ymin, y_text)



import numpy as np


def align_trees(tree_1, tree_2):
    """ Align two trees if possible.

    Args:
        tree_1 (dollo.trees.TreeNode): first tree to align
        tree_2 (dollo.trees.TreeNode): second tree to align

    Returns:
        list of tuple: matching between node labels in `tree_1` to node labels in `tree_2`

    Raises:
        ValueError: trees are not alignable

    Align trees if possible and return a mapping between node labels.
    Leaves are matched by leaf name.  Internal nodes are matched if they
    contain their set of descendent leaf names is equal.

    """

    if len(list(tree_1.nodes)) != len(list(tree_2.nodes)):
        raise ValueError('Trees not alignable')

    node_alignment = list()

    for node_1 in tree_1.nodes:

        leaf_set_1 = set([leaf.name for leaf in node_1.leaves])

        matched_node_2 = None
        for node_2 in tree_2.nodes:
            leaf_set_2 = set([leaf.name for leaf in node_2.leaves])
            if leaf_set_1 == leaf_set_2:
                matched_node_2 = node_2

        if matched_node_2 is None:
            raise ValueError('Trees not alignable')

        node_alignment.append((node_1.label, matched_node_2.label))

    return node_alignment


def simplify_tree(tree, node_labels):
    """ Simplify tree based on subset of nodes.

    Args:
        tree (dollo.TreeNode): tree to simplify
        node_labels (list): node labels for subset of nodes

    Returns:
        dollo.TreeNode: simplified trees with redundant nodes/branches removed

    """

    # Filter redundant branches
    while True:
        altered = False
        for node in tree.nodes:
            new_children = list()
            for child in node.children:
                if child.label not in node_labels:
                    new_children.extend(child.children)
                    altered = True
                else:
                    new_children.append(child)
            node.children = new_children
            if altered:
                break
        if not altered:
            break

    while True:
        if tree.label not in node_labels and len(tree.children) == 1:
            tree = tree.children[0]
        else:
            break

    return tree


def simplify_tree_by_genotype(tree, genotypes):
    """ Simplify tree based on given genotypes.

    Args:
        tree (dollo.TreeNode): tree to simplify
        genotypes (pandas.DataFrame): node table for genotypes

    Returns:
        dollo.TreeNode: simplified trees with redundant nodes/branches removed

    """

    # Add genotypes to nodes
    for node in tree.nodes:
        node.genotype = set(
            genotypes.loc[
                (genotypes['node'] == node.label) &
                (genotypes['ml_presence'] == 1),
                'cluster_id'
            ].values
        )

    # Filter redundant branches
    while True:
        altered = False
        for node in tree.nodes:
            new_children = list()
            for child in node.children:
                if node.genotype == child.genotype:
                    new_children.extend(child.children)
                    altered = True
                else:
                    new_children.append(child)
            node.children = new_children
            if altered:
                break
        if not altered:
            break

    while True:
        if len(tree.genotype) == 0:
            if len(tree.children) > 1:
                raise Exception('no ancestral genotype')
            tree = tree.children[0]
        else:
            break

    return tree


def classify_tree(tree):
    """ Classify tree type.

    Args:
        tree (dollo.TreeNode): tree to classify

    Returns:
        str: classification as 'pure', 'branched', 'chain'

    """

    if len(list(tree.nodes)) == 1:
        return 'pure'

    for node in tree.nodes:
        if len(node.children) > 1:
            return 'branched'

    return 'chain'


def uniform_transition(state, child_state):
    """ Uniform transition score

    Args:
        state (int): state of parent
        child_state (int): state of child

    Returns:
        int: score

    """

    if state == child_state:
        return 0
    else:
        return 1


def calculate_score_recursive(node, f_transition_score):
    """ Tree recursion for maximum parsimony.

    Args:
        node (dollo.TreeNode): node for recursive score calculation
        f_transition_score (callable): state transition function

    Node variables:
        state_scores (numpy.array): array of scores for each state
        state_backtrack (list of list): array of minimizing states for each parent state

    Assumes state_scores are valid at leaves.  At return, state_scores valid at
    all nodes, and state_backtrack valid at all non-root nodes.

    """

    if not node.is_leaf:

        node.state_scores = np.repeat(0., len(node.state_scores))

        for child in node.children:
            calculate_score_recursive(child, f_transition_score)
            child.state_backtrack = [[]] * len(node.state_scores)

        for state in xrange(len(node.state_scores)):

            for child in node.children:

                child_state_scores = np.repeat(0., len(node.state_scores))

                for child_state in xrange(0, len(node.state_scores)):

                    transition_score = f_transition_score(state, child_state)
                    child_state_scores[child_state] = transition_score + child.state_scores[child_state]

                min_score = np.amin(child_state_scores)

                child.state_backtrack[state] = np.where(min_score == child_state_scores)[0]
                node.state_scores[state] += min_score


def backtrack_state_recursive(node, state=None):
    """ Tree backtracking for maximum parsimony state.

    Args:
        node (dollo.TreeNode): node for recursive score calculation

    KwArgs:
        state (int): state assignment for node

    Node variables:
        state_scores (numpy.array): array of scores for each state
        state_backtrack (list of list): array of minimizing states for each parent state

    Assumes state_scores and state_backtrack are valid for all nodes.  At return,
    each node has a state variable representing state for one maximum parsimony
    solution.    

    """

    if state is None:
        node.state = np.argmin(node.state_scores)
    else:
        node.state = state

    for child in node.children:
        backtrack_state_recursive(child, child.state_backtrack[node.state][0])





'''
Position annotation functionality

Annotations for positional data such as SNVs and breakends.

Functions take pandas tables and add additional columns.
'''

import collections
import numpy as np
import pandas as pd

from wgs_analysis.algorithms.position import calculate_adjacent_density
from wgs_analysis.algorithms.position import calculate_adjacent_distance


def annotate_adjacent_density(df, stddev=5000):
    """ Add adjacent_density column

    Args:
        df(pandas.DataFrame): positions table

    Returns:
        pandas.DataFrame: original table with `adjacent_density` column

    Calculate the density of adjacent positions at each position using a gaussian
    kernal with standard deviation stddev.

    """

    positions = (
        df[['chrom', 'coord']]
        .drop_duplicates()
        .set_index(['chrom', 'coord'], drop=False)
        .sort_index())
    positions['coord'] = positions['coord'].astype(float)

    adjacent_density = (
        positions
        .groupby(level='chrom')['coord']
        .transform(calculate_adjacent_density, stddev))

    df.set_index(['chrom', 'coord'], inplace=True)
    df['adjacent_density'] = adjacent_density
    df.reset_index(inplace=True)

    return df


def annotate_adjacent_distance(df):
    """ Add adjacent_distance column

    Args:
        df(pandas.DataFrame): positions table

    Returns:
        pandas.DataFrame: original table with `adjacent_distance` column

    Calculate the closest distance to an adjacent position.
    """

    positions = (
        df[['chrom', 'coord']]
        .drop_duplicates()
        .set_index(['chrom', 'coord'], drop=False)
        .sort_index())
    positions['coord'] = positions['coord'].astype(float)

    adjacent_distance = (
        positions
        .groupby(level='chrom')['coord']
        .transform(lambda pos: calculate_adjacent_distance(pos.values)))

    df.set_index(['chrom', 'coord'], inplace=True)
    df['adjacent_distance'] = adjacent_distance
    df.reset_index(inplace=True)

    return df




'''
# Position plots

Plotting of positions in chromosome plots using stem plots.
'''

import numpy as np
import matplotlib.pyplot as plt
import seaborn

def plot_positions(ax, positions, chrom, site):

    positions = positions[positions['chrom'] == chrom]
    positions = positions[positions['site_id'] == site]
    positions = positions[positions['alt_counts'] > 0]
    for coord in positions['coord'].unique():
        ax.plot([coord, coord], [-0.2, 0.2], color='#146414', linewidth=0.75)


def plot_breakends(ax, brks, site_id, chromosome, start=None, end=None, xlim=None, ylim=None, fcolor=None, lw=None):

    if xlim is None:
        xlim = ax.get_xlim()
    xrng = xlim[1] - xlim[0]

    if ylim is None:
        ylim = ax.get_ylim()
    yrng = ylim[1] - ylim[0]

    arrow_length = 0.01 * xrng
    
    brks = brks[(brks['chrom'] == chromosome)]

    if start is not None:
        brks = brks[(brks['coord'] > start)]
    
    if end is not None:
        brks = brks[(brks['coord'] < end)]

    brks = brks[(brks['site_id'] == site_id)]
    brks = brks[(brks['num_span'] > 0)]

    head_width = yrng / 30.
    head_length = xrng / 100.

    for strand, positions in brks.groupby('strand')['coord']:
        xpos = positions.values
        ypos = np.random.uniform(ylim[0] + 0.75*yrng,
                                 ylim[0] + 0.95*yrng,
                                 size=len(positions))
        offset = (arrow_length, -arrow_length)[strand == '+']
        for x, y in zip(xpos, ypos):
            color = 'grey'
            if fcolor is not None:
                color = fcolor(chromosome, strand, x)
            ax.arrow(x, y, offset, 0, color=color, lw=lw, alpha=1.0, head_width=head_width, head_length=head_length)
            ax.plot([x, x], [-100, 100], color=color, lw=lw, alpha=1.0)
    



import os
import warnings
import random
import pandas as pd

import wgs_analysis.experiments.design_primers as design_primers
import wgs_analysis.experiments.design_snv_primers as design_snv_primers
import wgs_analysis.experiments.design_break_primers as design_break_primers
import blat_server



def design_categories(selection, designer, genome_fasta, seed=2014, primer3_params=None, **kwargs):
    """ Design primers in groups with group specific requirements

    Args:
        selection(pandas.DataFrame) : selected variants table, see format below
        designer(callable) : design function, either design_snv_primers.design or design_break_primers.design
        genome_fasta(str) : genome fasta filename

    KwArgs:
        seed(int) : random seed for shuffling
        primer3_params(dict) : Additional Primer 3 Parameters
        kwargs : additional arguments to design function

    Returns:
        pandas.DataFrame of primers

    The selection table requires the following columns:
     * seq_id : unique identifier for the SNV/breakpoint
     * category : reason the SNV/breakpoint was selected
     * max_primer_stage : maximum stage at which primer selection will fail, -1 to try all stages
     * subset_count : number of SNVs/breakpoints to design from this category

    """

    random.seed(2014)

    primer_table = list()

    excluded_events = set()

    # Design SNV primers for each category
    for category, variants in selection.groupby('category'):

        # Remove variants for which we have already designed primers
        variants = variants[~variants['seq_id'].isin(excluded_events)]

        # Check if any variants remain to be designed for
        if len(variants.index) == 0:
            continue

        subset_count = variants['subset_count'].iloc[0]
        max_primer_stage = variants['max_primer_stage'].iloc[0]
        
        # Shuffle variants within this category
        shuffled_index = list(variants.index.values)
        random.shuffle(shuffled_index)
        variants = variants.ix[shuffled_index]

        # Design primers
        primers = designer(genome_fasta, variants, max_stage=max_primer_stage, max_primers=subset_count, primer3_params=primer3_params, **kwargs)

        # Warn if requirements not met
        if subset_count >= 0:
            if len(primers.index) < subset_count:
                warnings.warn('insufficient variants for category ' + category)
        else:
            for seq_id in variants['seq_id'].values:
                if seq_id not in primers['seq_id'].values:
                    warnings.warn('no primers for variant ' + str(seq_id) + ' in category ' + category)

        primer_table.append(primers)

        excluded_events.update(primers['seq_id'].values)

    primer_table = pd.concat(primer_table, ignore_index=True).drop_duplicates()

    for seq_id, primers in primer_table.groupby('seq_id'):
        if len(primers) > 1:
            warnings.warn('designed ' + str(len(primers)) + ' primers for sequence ' + str(seq_id) + ' in categories ' + ', '.join(primers['category']))

    return primer_table



def print_design_stats(**kwargs):
    """
    Print the number of events for which primers will be designed

    Args:
        kwargs : pandas.DataFrame tables of events with the format as described for design_seqval

    """

    primer_pair_count = 0

    for event_type, events in kwargs.iteritems():

        if events is None:
            continue

        for category, category_events in events.groupby('category'):

            subset_count = category_events['subset_count'].iloc[0]
            if subset_count < 0:
                subset_count = len(category_events.index)

            total_count = len(category_events.index)

            primer_pair_count += subset_count

            print 'designing {0}/{1} for {2} {3}'.format(subset_count, total_count, category, event_type)

    print 'designing {0} primer pairs'.format(primer_pair_count)



def print_experiment_stats(**kwargs):
    """
    Print the number of events for which primers have been designed

    Args:
        kwargs : pandas.DataFrame tables of events with the format as described for design_seqval

    """

    primer_pair_count = 0

    for event_type, events in kwargs.iteritems():

        if events is None:
            continue

        for category, category_events in events.groupby('category'):
            total_count = len(category_events.index)

            primer_pair_count += total_count

            print 'designed {0} for {1} {2}'.format(total_count, category, event_type)

    print 'designing {0} primer pairs'.format(primer_pair_count)



def design_experiment(patient_id, experiment_id, experiment_type, genome_fasta, variant_vcfs, snvs=None, breakpoints=None, primer3_params=None):
    """
    Design primers for both SNVs and breakpoints and write to the appropriate location in the
    meta_data directory.

    Args:
        patient_id(str) : patient identifier
        experiment_id(str) : experiment identifier to distinguish different experiments per patient
        experiment_type(str) : type of experiment, either `seqval` or `singlecell`
        genome_fasta(str) : genome fasta filename
        variant_vcfs(list) : list of vcfs containing variants to avoid

    KwArgs:
        snvs(pandas.DataFrame) : SNVs selected for primer design
        breakpoints(pandas.DataFrame) : Breakpoints selected for primer design
        primer3_params(dict) : Additional Primer 3 Parameters

    Breakpoints and SNVs are given as pandas tables with the following required columns:
     * seq_id : unique identifier for the SNV/breakpoint
     * category : reason the SNV/breakpoint was selected
     * max_primer_stage : maximum stage at which primer selection will fail, -1 to try all stages
     * subset_count : number of SNVs/breakpoints to design from this category

    Additionally, SNVs require the following columns:
     * chrom : chromosome of SNV
     * coord : coordinate of the SNV
     * ref : reference base of the SNV
     * alt : variant base of the SNV

    Additionally, breakpoint tables are 2 rows per breakpoint, one for each breakend, and require the
    following columns:
     * chrom : pair of chromosomes, comma separated
     * strand : pair of strands, comma separated
     * coord : pair of coordinates, comma separated

    """

    # Check args

    if experiment_type not in ('seqval', 'singlecell'):
        raise ValueError('unsupported experiment type')

    required_snv_columns = set(['seq_id', 'category', 'max_primer_stage', 'subset_count',
                                'chrom', 'coord', 'ref', 'alt'])

    required_breakpoint_columns = set(['seq_id', 'category', 'max_primer_stage', 'subset_count',
                                       'chrom', 'strand', 'coord'])

    if snvs is not None and not required_snv_columns.issubset(snvs.columns.values):
        raise ValueError('missing columns from snv table')

    if breakpoints is not None and not required_breakpoint_columns.issubset(breakpoints.columns.values):
        raise ValueError('missing columns from breakpoint table')


    # Design primers

    print_design_stats(snvs=snvs, breakpoints=breakpoints)

    primer_table = list()

    with blat_server.BlatServer(genome_fasta):

        if snvs is not None:

            print 'designing primers for SNVs'

            snv_primers = design_categories(snvs, design_snv_primers.design,
                genome_fasta, variant_vcfs=variant_vcfs,
                requirements_filename=design_snv_primers.get_requirements_filename(experiment_type),
                primer3_params=primer3_params)

            snv_primers['variant_type'] = 'snv'
            snv_primers['strand'] = '+'

            primer_table.append(snv_primers)

        if breakpoints is not None:

            print 'designing primers for breakpoints'

            breakpoint_primers = design_categories(breakpoints, design_break_primers.design,
                genome_fasta, variant_vcfs=variant_vcfs,
                requirements_filename=design_break_primers.get_requirements_filename(experiment_type),
                primer3_params=primer3_params)

            breakpoint_primers['variant_type'] = 'breakpoint'
            breakpoint_primers['ref'] = 'NA'
            breakpoint_primers['alt'] = 'NA'

            primer_table.append(breakpoint_primers)


    primer_table = pd.concat(primer_table, ignore_index=True)


    # Merge duplicates, created comma delimted list in category column

    def merge_duplicates_pairs(df):
        if len(df.index) == 1:
            return df
        else:
            categories = ','.join(df['category'])
            df = df.sort('stage')
            df = df.iloc[0:1]
            df['category'] = categories
            return df

    primer_table = primer_table.groupby('seq_id').apply(merge_duplicates_pairs)

    primer_table['patient_id'] = patient_id

    print 'designed {} primers'.format(len(primer_table.index))

    print_experiment_stats(
        snvs=primer_table[primer_table['variant_type'] == 'snv'],
        breakpoints=primer_table[primer_table['variant_type'] == 'breakpoint'])

    return primer_table



def create_snv_seq_id(snvs):
    return snvs.apply(lambda row: '{0}:{1}'.format(row['chrom'], row['coord']), axis=1)



def create_snvs_table(selection, snvs):
    """
    Create a breakpoint table compatible with design_seqval

    Args:
        selection(pandas.DataFrame) : selected snvs table, chrom and coord columns required
        snvs(pandas.DataFrame) : full SNVs table

    Returns:
        pandas.DataFrame selection breakpoints with all columns from selection and the following additional columns:
         * seq_id
         * ref
         * alt

    """

    selection_columns = list(selection.columns.values)
    additional_columns = ['ref', 'alt']

    snvs = snvs.merge(selection, on=['chrom', 'coord'], how='inner')
    snvs = snvs[selection_columns + additional_columns].drop_duplicates()
    snvs['seq_id'] = create_snv_seq_id(snvs)

    return snvs



def create_breakpoint_seq_id(snvs):
    return snvs.apply(lambda row: row['prediction_id'], axis=1)



def create_breakpoints_table(selection, breakpoints):
    """
    Create a breakpoint table compatible with design_seqval

    Args:
        selection(pandas.DataFrame) : selected breakpoints table, prediction_id column required
        breakpoints(pandas.DataFrame) : full breakpoints table

    Returns:
        pandas.DataFrame selection breakpoints with all columns from selection and the following additional columns:
         * seq_id
         * chrom : pair of chromosomes, comma separated
         * strand : pair of strands, comma separated
         * coord : pair of coordinates, comma separated
  
    """

    breakpoints = breakpoints.merge(selection[['prediction_id']].drop_duplicates(), how='inner').set_index('prediction_id')

    def calculate_chrom(row):
        return '{0},{1}'.format(row['chromosome_1'], row['chromosome_2'])
        
    def calculate_strand(row):
        return '{0},{1}'.format(row['strand_1'], row['strand_2'])
        
    def calculate_coord(row):
        return '{0},{1}'.format(row['position_1'], row['position_2'])

    chrom = breakpoints.apply(calculate_chrom, axis=1)
    strand = breakpoints.apply(calculate_strand, axis=1)
    coord = breakpoints.apply(calculate_coord, axis=1)

    breakpoints = pd.DataFrame({'chrom':chrom, 'strand':strand, 'coord':coord}).reset_index()

    breakpoints = breakpoints.merge(selection, on='prediction_id', how='inner')

    breakpoints['seq_id'] = create_breakpoint_seq_id(breakpoints)
    breakpoints = breakpoints.drop('prediction_id', axis=1)

    return breakpoints



def calculate_seq_id_category_map(primer_table):
    """
    Given a primer table calculate a mapping between seq_id and categories

    Args:
        primer_table(pandas.DataFrame) : data frame with seq_id and category as mandatory columns

    Returns:
        pandas.DataFrame with seq_id and category columns

    We are storing categories as comma separated lists in the primer table.  This helper function extracts
    a one to many mapping between seq_ids and individual categories. 

    """

    seq_id_category = list()
    for seq_id, categories in primer_table[['seq_id', 'category']].values:
        for category in categories.split(','):
            seq_id_category.append((seq_id, category))
            
    return pd.DataFrame(seq_id_category, columns=['seq_id', 'category'])





'''
# Tree plotting

'''
from matplotlib.patches import Rectangle as Rectangle

import dollo
import itertools
import numpy as np
import seaborn

import wgs_analysis.plots.utils


def add_depths(tree, depth):
    """ Add plot depths (y-axis) for noninformative branch lengths

    Args:
        tree (dollo.trees.TreeNode): tree to add depths
        depth (float): starting depth of subtree rooted at `tree`

    """

    tree.depth = depth
    for child in tree.children:
        add_depths(child, depth+1)


def add_depths_brlen(tree, branch_length_attr, parent_depth=0.):
    """ Add plot depths (y-axis) based on branch lengths.

    Args:
        tree (dollo.trees.TreeNode): tree to add depths
        branch_length_attr (str): tree node attribute to use as branch length for node/parent branch
        parent_depth (float): starting depth of subtree rooted at `tree`

    """

    tree.depth = parent_depth + getattr(tree, branch_length_attr)
    for child in tree.children:
        add_depths_brlen(child, branch_length_attr, tree.depth)
        

def add_positions(tree, leaf_cnt):
    """ Add plot positions (x-axis) for each node.

    Args:
        tree (dollo.trees.TreeNode): tree to add positions

    Leaves are evenly spaced.  Internal nodes are placed at the midpoint
    between the left and right-most positions of the children.

    """

    if len(tree.children) > 0:
        for child in tree.children:
            add_positions(child, leaf_cnt)
        posns = [child.position for child in tree.children]
        tree.position = (max(posns) + min(posns)) / 2.
    else:
        tree.position = float(next(leaf_cnt))


def plot_tree(ax, tree, branch_length_attr=None, leaf_name_attr=None, extend_leaf_branch=False, landscape=True, flip=False, **kwargs):
    """ Plot a tree.

    Args:
        ax (matplotlib.axes.Axes): plot axes
        tree (dollo.trees.TreeNode): tree to plot

    Kwargs:
        branch_length_attr (str): node attribute to use as branch length
        leaf_name_attr (str): node attribute to use as leaf name
        extend_leaf_branch (bool): extend leafs to axis when not using branch lengths
        landscape (bool): plot tree in landscape mode (evolution axis as x axis)
        flip (bool): flip the evolution axis
        **kwargs: additional kwargs to plot

    """

    if extend_leaf_branch and branch_length_attr is not None:
        raise ValueError('cannot extend leaf branches when given branch lengths')

    if 'linestyle' not in kwargs:
        kwargs['linestyle'] = '-'

    if 'color' not in kwargs:
        kwargs['color'] = 'k'

    tree.add_parent()
    
    if branch_length_attr is not None:
        add_depths_brlen(tree, branch_length_attr)
    else:
        add_depths(tree, 1)

    max_depth = max([node.depth for node in tree.nodes])

    if extend_leaf_branch:
        for node in tree.nodes:
            if node.is_leaf:
                node.depth = max_depth

    add_positions(tree, itertools.count())

    # Assign x,y positions to nodes
    for node in tree.nodes:
        if landscape:
            node.x = node.depth
            node.y = node.position
        else:
            node.x = node.position
            node.y = node.depth

    # Assign branch line segments to nodes
    for node in tree.nodes:
        parent_x, parent_y = 0, 0
        if node.parent is not None:
            parent_x, parent_y = node.parent.x, node.parent.y
        if landscape:
            node.branch = ((node.x, parent_x), (node.y, node.y))
        else:
            node.branch = ((node.x, node.x), (node.y, parent_y))

    # Create a collection of lateral edges to link node branches
    links = list()
    for node in tree.nodes:
        if node.is_leaf:
            continue
        brx = [child.branch[0][1] for child in node.children]
        bry = [child.branch[1][1] for child in node.children]
        links.append(((min(brx), max(brx)), (min(bry), max(bry))))

    # Plot edges
    for node in tree.nodes:
        node.branch_line = ax.plot(node.branch[0], node.branch[1], **kwargs)[0]

    # Plot lateral links between edges
    for link in links:
        ax.plot(link[0], link[1], **kwargs)

    xs = [node.x for node in tree.nodes] + [0]
    ys = [node.y for node in tree.nodes] + [0]

    wgs_analysis.plots.utils.set_xlim_filter_ticks(ax, min(xs), max(xs))
    wgs_analysis.plots.utils.set_ylim_filter_ticks(ax, min(ys), max(ys))

    ax.grid(False)

    if flip:
        if landscape:
            ax.invert_xaxis()
        else:
            ax.invert_yaxis()

    # Initialize all spines to invisible
    for spine in ax.spines:
        ax.spines[spine].set_visible(False)
    ax.xaxis.set_tick_params(which='both', bottom=False, top=False, labelbottom=False, labeltop=False)
    ax.yaxis.set_tick_params(which='both', left=False, right=False, labelleft=False, labelright=False)

    # Set spine and ticks to visible for x/y axis left/right/top/bottom
    def setup_axis(axis, spine):
        ax.spines[spine].set_visible(True)
        ax.spines[spine].set_position(('outward', 10))
        axis.set_tick_params(which='both', **{spine:True, 'label'+spine:True})

    # Setup axis ticks and spine visibility based on orientation
    if landscape:
        if branch_length_attr is not None:
            setup_axis(ax.xaxis, 'bottom')
        if flip:
            setup_axis(ax.yaxis, 'left')
        else:
            setup_axis(ax.yaxis, 'right')
    else:
        if branch_length_attr is not None:
            setup_axis(ax.yaxis, 'left')
        if flip:
            setup_axis(ax.xaxis, 'bottom')
        else:
            setup_axis(ax.xaxis, 'top')

    # Shift plot limits to allow for full visibility of lines
    wgs_analysis.plots.utils.shift_lims(ax, 2, 2)

    # Trim spines to ticks for both axes
    wgs_analysis.plots.utils.trim_spines_to_ticks(ax)

    # Set leaf names if we have an attribute to query from leaf nodes
    leaf_ticks = list(range(len(list(tree.leaves))))
    if leaf_name_attr is not None:
        leaf_names = [getattr(leaf, leaf_name_attr) for leaf in tree.leaves]
        if landscape:
            ax.set_yticks(leaf_ticks)
            ax.set_yticklabels(leaf_names)
        else:
            ax.set_xticks(leaf_ticks)
            ax.set_xticklabels(leaf_names)


def plot_tree_nodes(ax, tree, node_color=None, node_label=False, node_attr=None, node_marker_size=6):
    """ Plot a tree.

    Args:
        ax (matplotlib.axes.Axes): plot axes
        tree (dollo.trees.TreeNode): tree to plot

    Kwargs:
        node_color (color or list of colors): single color for all nodes or color of each node by label
        node_label (bool): whether to draw node labels
        node_attr (str): node attribute to plot
        node_marker_size (int): size of circular node marker

    """

    for node in tree.nodes:

        if node_color is not None or node_label:

            if node_color is None:
                color = 'white'
            elif isinstance(node_color, basestring):
                color = node_color
            else:
                color =  node_color[node.label]

            ax.plot(node.x, node.y, 'bo', color=color, zorder=2, fillstyle='full', markersize=node_marker_size)

        node_text = None
        if node_label:
            node_text = str(node.label)
        elif node_attr is not None:
            node_text = str(getattr(node, node_attr))

        if node_text is not None:
            ax.text(node.x, node.y, node_text, ha='center', va='center', fontsize='small')


def plot_genotype_nodes(ax, tree, genotype_attr=None, genotype_colors=None, width=0.1, height=0.1):
    """ Plot a genotype tree.

    Args:
        ax (matplotlib.axes.Axes): plot axes
        tree (dollo.trees.TreeNode): tree to plot

    Kwargs:
        genotype_attr (str): attribute for presence absence vector on each node
        genotype_colors (list of colors): color for each variant in the genotype
        width (float): width of genotype boxes
        height (float): height of genotype boxes

    The tree must have an array of 0/1 indicators at each node denoting presence/absence
    of the variant for the genotype of each node.

    """

    for node in tree.nodes:

        offset_x = width * float(len(node.genotype)) / 2.
        offset_y = -height * 0.5

        for idx, present in enumerate(node.genotype):

            widget_x =node.x + width * idx - offset_x
            widget_y = node.y + offset_y

            color = 'white'

            if present == 1:
                if genotype_colors is None:
                    color = 'k'
                else:
                    color = genotype_colors[idx]

            ax.add_patch(Rectangle((widget_x, widget_y), width, height,
                facecolor=color, edgecolor='.75', linewidth=0.25, zorder=10))

        ax.add_patch(Rectangle((node.x - offset_x, node.y + offset_y), 2 * offset_x, height,
            facecolor=(0,0,0,0), edgecolor='k', linewidth=1., zorder=10))


def plot_branch_events(ax, tree, genotype_table, event_colors=None):
    """ Plot origin and loss on branches.

    Args:
        ax (matplotlib.axes.Axes): plot axes
        tree (dollo.trees.TreeNode): tree to plot
        genotype_table (pandas.DataFrame): table of genotypes per node

    KwArgs:
        event_colors (dict): event colors keyed by event id

    The genotype table should have node labels as row indices and event ids (ids of 
    snv/breakpoint clusters) as column labels.  Each 0/1 entry represents the
    presence or absence of the event for that node.

    """

    num_events = len(genotype_table.columns)

    if event_colors is None:
        if num_events > 8:
            palette = seaborn.color_palette('husl', num_events)
        else:
            palette = seaborn.color_palette('Set2', num_events)
        event_colors = dict(zip(genotype_table.columns.values, palette))

    for node in tree.nodes:

        markers = list()

        for event_id in genotype_table.columns.values:

            color = event_colors[event_id]

            node_present = genotype_table.loc[node.label, event_id]

            parent_present = 0
            if node.parent is not None:
                parent_present = genotype_table.loc[node.parent.label, event_id]

            if parent_present == 0 and node_present == 1:
                markers.append({'marker':'o', 'color':color, 'zorder':100, 'facecolor':'white', 's':50, 'lw':3})
            elif parent_present == 1 and node_present == 0:
                markers.append({'marker':'x', 'color':color, 'zorder':100, 's':50, 'lw':2})

        if len(markers) == 0:
            continue

        positions = np.linspace(0, 1, len(markers) + 2)[1:-1]

        for pos, marker in zip(positions, markers):
            x = pos * node.branch[0][0] + (1 - pos) * node.branch[0][1]
            y = pos * node.branch[1][0] + (1 - pos) * node.branch[1][1]
            ax.scatter(x, y, **marker)


def get_mutation_order(tree, nodes):
    """ Get mutation order placing ancestral first.

    Args:
        tree (dollo.trees.TreeNode): tree to plot
        nodes (pandas.DataFrame): nodes table

    Returns:
        list: event_ids from nodes table ordered by ancestrality

    """

    add_depths(tree, 0)
    add_positions(tree, itertools.count())

    for node in tree.nodes:
        nodes.loc[nodes['node'] == node.label, 'plot_depth'] = node.depth
        nodes.loc[nodes['node'] == node.label, 'plot_position'] = node.position

    event_origin = nodes[nodes['ml_origin'] == 1]
    event_origin = event_origin.sort(['plot_depth', 'plot_position'])
    event_order = event_origin['category'].values

    return event_order


def add_internal_node_leaves(tree):
    """ Add leaves to represent internal nodes.

    Args:
        tree (dollo.trees.TreeNode): tree to augment

    Returns:
        dollo.trees.TreeNode: tree with additional leaves

    Additional leaf nodes will have the same label as their internal node parent.

    """

    for node in list(tree.nodes):
        if not node.is_leaf:
            leaf = dollo.trees.TreeNode(None)
            leaf.label = node.label
            node.children.append(leaf)

    return tree






'''
# Design primers module

Provides functionality for designing primers based on input sequences provided in a pandas table.  A blat server must be
running, and the convencience class BlatServer provides a context manager based server for a given genome.  Runs primer3
and isPCR to obtain primers, and applies constraints to filter primer pairs.  Progressively relaxes constraints if
requested.

'''

import argparse
import csv
import collections
import sys
import os
import tempfile
import time
import math
import subprocess
import numpy as np
import pandas as pd
import distutils.spawn

import design_utils
import blat_server


primer3_core_ex = distutils.spawn.find_executable('primer3_core')
if primer3_core_ex is None:
    raise Exception('primer3_core not found')

primer3_dir = os.path.dirname(os.path.realpath(primer3_core_ex))

default_primer3_parameters = {
    'PRIMER_MIN_SIZE': 18,
    'PRIMER_MAX_SIZE': 26,
    'PRIMER_NUM_NS_ACCEPTED': 0,
    'PRIMER_FILE_FLAG': 0,
    'PRIMER_PICK_INTERNAL_OLIGO': 0,
    'PRIMER_MIN_TM': 58.0,
    'PRIMER_OPT_TM': 60.0,
    'PRIMER_MAX_TM': 62.0,
    'PRIMER_MAX_POLY_X': 4,
    'PRIMER_THERMODYNAMIC_TEMPLATE_ALIGNMENT': 1,
    'PRIMER_THERMODYNAMIC_PARAMETERS_PATH': '{primer3_dir}/primer3_config/'.format(primer3_dir=primer3_dir),
    'PRIMER_MISPRIMING_LIBRARY': '{primer3_dir}/humrep_and_simple.txt'.format(primer3_dir=primer3_dir),
}


def run_primer3(sequence, req, params=None):
    '''
    Run primer 3 and obtain a list of potential primers.
    '''
    target_start = sequence.index('[')
    target_size = sequence.index(']') - sequence.index('[') - 1
    sequence = sequence.replace('[', '').replace(']', '')

    req_target_start = target_start - int(req['target_buffer'])
    req_target_size = target_size + 2 * int(req['target_buffer'])

    primers = list()

    primer3_parameters = default_primer3_parameters.copy()
    primer3_parameters['PRIMER_PRODUCT_SIZE_RANGE'] = '{0}-{1}'.format(req['min_product_size'], req['max_product_size'])
    primer3_parameters['PRIMER_OPT_SIZE'] = '{0}'.format(req['opt_size'])
    if req['gc_clamp'] >= 0:
        primer3_parameters['PRIMER_GC_CLAMP'] = '{0}'.format(req['gc_clamp'])

    if params is not None:
        primer3_parameters.update(params)

    primer3_proc = subprocess.Popen(['primer3_core'], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    primer3_proc.stdin.write('SEQUENCE_TEMPLATE={0}\n'.format(sequence))
    primer3_proc.stdin.write('SEQUENCE_TARGET={0},{1}\n'.format(req_target_start, req_target_size))
    for key, value in primer3_parameters.iteritems():
        primer3_proc.stdin.write('{}={}\n'.format(key, value))
    primer3_proc.stdin.write('=\n')
    primer3_proc.stdin.close()

    pr3 = dict(line.rstrip().split('=') for line in primer3_proc.stdout)
    primer3_proc.stdout.close()

    if primer3_proc.wait() != 0:
        sys.stderr.write('primer3 errors\n')
        sys.exit(1)

    for primer_id in range(0, 5):
        if 'PRIMER_LEFT_{0}_SEQUENCE'.format(primer_id) in pr3:
            primer_info = collections.OrderedDict()
            primer_info['primer_id'] = primer_id
            primer_info['sequence'] = sequence
            primer_info['left_primer'] = pr3['PRIMER_LEFT_{0}_SEQUENCE'.format(primer_id)]
            primer_info['right_primer'] = pr3['PRIMER_RIGHT_{0}_SEQUENCE'.format(primer_id)]
            primer_info['left_primer_length'] = pr3['PRIMER_LEFT_{0}'.format(primer_id)].split(',')[1]
            primer_info['right_primer_length'] = pr3['PRIMER_RIGHT_{0}'.format(primer_id)].split(',')[1]
            primer_info['left_tm'] = pr3['PRIMER_LEFT_{0}_TM'.format(primer_id)]
            primer_info['right_tm'] = pr3['PRIMER_RIGHT_{0}_TM'.format(primer_id)]
            primer_info['product_size'] = pr3['PRIMER_PAIR_{0}_PRODUCT_SIZE'.format(primer_id)]
            primer_info['target_buffer'] = req['target_buffer']
            primer_info['gc_clamp'] = req['gc_clamp']
            primer_info['stage'] = req['stage']
            primers.append(primer_info)

    primers = pd.DataFrame(primers)

    return primers


def run_insilico_pcr(primers):
    '''
    Use gfPcr to count the number of potential products in the human genome.
    Add the product count to the list of dictionaries in primers.
    '''
    products = list()

    with design_utils.TempDirectory() as temps_dir:

        temp_input_filename = os.path.join(temps_dir, 'primers.tsv')
        temp_output_filename = os.path.join(temps_dir, 'products.fa')

        with open(temp_input_filename, 'w') as temp_input_file:
            for idx, primer in primers.iterrows():
                temp_input_file.write('{0}\t{1}\t{2}\t4000\n'.format(primer['primer_id'], primer['left_primer'], primer['right_primer']))

        is_pcr_proc = subprocess.check_call(['gfPcr', 'localhost', '8899', '/', temp_input_filename, temp_output_filename, '-out=psl'], stdout=subprocess.PIPE)

        with open(temp_output_filename, 'r') as temp_output_file:
            for line in temp_output_file:
                row = line.split()
                primer_id = int(row[9])
                chrom = row[13]
                start = int(row[15]) + 1
                end = int(row[16])
                products.append({'primer_id':primer_id, 'chrom':chrom, 'start':start, 'end':end})

    products = pd.DataFrame(products)

    if len(products.index) == 0:
        products = pd.DataFrame(columns=['primer_id', 'chrom', 'start', 'end'])

    return products


def run_alignment(primers):
    '''
    Use gfServer to count the number of alignments of each primer in the human genome.
    Add the alignment count to the list of dictionaries in primers.
    '''
    alignment_infos = list()

    with design_utils.TempDirectory() as temps_dir:

        temp_fasta_filename = os.path.join(temps_dir, 'primers.fa')
        temp_psl_filename = os.path.join(temps_dir, 'alignments.psl')

        with open(temp_fasta_filename, 'w') as temp_fasta_file:
            for idx, primer in primers.iterrows():
                temp_fasta_file.write('>{0}_left\n{1}\n'.format(idx, primer['left_primer']))
                temp_fasta_file.write('>{0}_right\n{1}\n'.format(idx, primer['right_primer']))

        subprocess.check_call(['gfClient', 'localhost', '8899', '/', temp_fasta_filename, temp_psl_filename, '-minScore=0', '-minIdentity=0', '-nohead'], stdout=subprocess.PIPE)

        with open(temp_psl_filename, 'r') as temp_psl_file:
            for row in csv.reader(temp_psl_file, delimiter='\t'):
                matches = row[0]
                length = row[10]
                if matches != length:
                    continue
                idx, side = row[9].split('_')
                idx = int(idx)
                primer_id = primers.loc[idx, 'primer_id']
                alignment_infos.append({'primer_id':primer_id, side+'_alignment_count':1})

    alignment_infos = pd.DataFrame(alignment_infos)

    alignment_infos = alignment_infos.reindex(columns=['primer_id', 'left_alignment_count', 'right_alignment_count'])

    alignment_infos = alignment_infos.fillna(0)
    alignment_infos['left_alignment_count'] = alignment_infos['left_alignment_count'].astype(int)
    alignment_infos['right_alignment_count'] = alignment_infos['right_alignment_count'].astype(int)

    alignment_infos = alignment_infos.groupby('primer_id').sum().reset_index()

    return alignment_infos


def pick_primers(requirements, seq_data, design_callback=None, max_stage=-1, max_primers=-1, primer3_params=None):
    '''
    Design primers, iteratively relaxing design requirements
    '''
    primer_table = list()

    for idx, (seq_id, sequence) in enumerate(seq_data):

        for req_idx, req_row in requirements.iterrows():

            if max_stage >= 0 and req_row['stage'] > max_stage:
                continue

            primers = run_primer3(sequence, req_row, params=primer3_params)
            primers['seq_id'] = seq_id

            if len(primers.index) == 0:
                continue

            def check_primer_uniqueness(row):
                if row['sequence'].count(row['left_primer']) != 1:
                    return False
                if row['sequence'].count(design_utils.reverse_complement(row['right_primer'])) != 1:
                    return False
                return True

            primers = primers[primers.apply(check_primer_uniqueness, axis=1)]

            if len(primers.index) == 0:
                continue

            products = run_insilico_pcr(primers)
            products['seq_id'] = seq_id

            primers.set_index('primer_id', inplace=True)
            primers['product_count'] = products.groupby('primer_id').size()
            primers['product_count'] = primers['product_count'].fillna(0).astype(int)
            primers.reset_index(inplace=True)

            alignment_info = run_alignment(primers)
            primers = primers.merge(alignment_info, on='primer_id').fillna(0)

            if req_row['min_product_count'] >= 0:
                primers = primers[primers['product_count'] >= req_row['min_product_count']]

            if req_row['max_product_count'] >= 0:
                primers = primers[primers['product_count'] <= req_row['max_product_count']]

            if req_row['max_alignment_count'] >= 0:
                primers = primers[primers['left_alignment_count'] <= req_row['max_alignment_count']]
                primers = primers[primers['right_alignment_count'] <= req_row['max_alignment_count']]

            if design_callback is not None:
                primers, products = design_callback(primers, products)

            if len(primers.index) > 0:
                primer_table.append(primers.iloc[0:1])
                sys.stderr.write('+')
                break
            else:
                sys.stderr.write('.')

        if max_primers >= 0 and len(primer_table) >= max_primers:
            break

    sys.stderr.write('\n')

    if len(primer_table) == 0:
        return pd.DataFrame(columns=[
            'primer_id', 'gc_clamp', 'left_primer', 'left_primer_length',
            'left_tm', 'product_size', 'right_primer', 'right_primer_length',
            'right_tm', 'sequence', 'stage', 'target_buffer', 'seq_id',
            'product_count', 'left_alignment_count', 'right_alignment_count',
            'product_start', 'product_end'])

    primer_table = pd.concat(primer_table, ignore_index=True)

    return primer_table


def design(requirements_filename, sequences, design_callback=None, max_stage=-1, max_primers=-1, primer3_params=None):

    requirements = pd.read_csv(requirements_filename, sep='\t')
    primer_table = pick_primers(requirements, sequences, design_callback=design_callback, max_stage=max_stage, max_primers=max_primers, primer3_params=primer3_params)
    return primer_table


if __name__ == '__main__':

    argparser = argparse.ArgumentParser()
    argparser.add_argument('genome', help='genome fasta filename')
    argparser.add_argument('requirements', help='design requirements table')
    argparser.add_argument('sequences', help='sequences fasta filename')
    argparser.add_argument('primers', help='primer table filename')
    args = argparser.parse_args()

    with blat_server.BlatServer(args.genome), open(args.sequences, 'r') as sequences_file:
        primer_table = design(args.requirements, design_utils.read_fasta(sequences_file))
        primer_table.to_csv(args.primers, sep='\t', index=False, na_rep='NA')



'''
# Design primers utils

Provides convencience functions used by primer design modules.

'''


import os
import shutil
import subprocess
import string
import uuid


def get_region_string(chrom, start, end):
    return '{0}:{1}-{2}'.format(chrom, int(start), int(end))

def fetch_sequence(genome, chrom, start, end):
    faidx_result = str(subprocess.check_output(['samtools', 'faidx', genome, get_region_string(chrom, start, end)]))
    seq_id, sequence = next(read_fasta(faidx_result.splitlines()))
    return sequence

def fetch_variants(vcf_filename, chrom, start, end):
    for line in subprocess.check_output(['tabix', vcf_filename, get_region_string(chrom, start, end)]).splitlines():
        if line == '':
            continue
        row = line.split()
        assert chrom == row[0]
        coord = int(row[1])
        ref = row[3]
        alt = row[4]
        length = len(ref)
        variant_start = max(start, coord)
        variant_end = min(end, coord + length - 1)
        yield variant_start, variant_end - variant_start + 1

def reverse_complement(sequence):
    return sequence[::-1].translate(string.maketrans('ACTGactg','TGACtgac'))

def read_fasta(fasta):
    ''' Read sequences from an open fasta file. Yield name, sequence pairs '''
    name = None
    sequences = []
    for line in fasta:
        line = line.rstrip()
        if len(line) == 0:
            continue
        if line[0] == '>':
            if name is not None:
                yield (name, ''.join(sequences))
            name = line[1:]
            sequences = []
        else:
            sequences.append(line)
    if name is not None:
        yield (name, ''.join(sequences))

def create_sequence(genome_fasta, chrom, start, end, variant_vcfs=[]):
    '''
    Create sequences spanning a breakpoint, and mask variants with N
    '''
    sequence = list(fetch_sequence(genome_fasta, chrom, start, end))

    for var_vcf in variant_vcfs:
        for var_start, var_length in fetch_variants(var_vcf, chrom, start, end):
            local_var_start = max(0, var_start - start)
            sequence[local_var_start:local_var_start+var_length] = ['N'] * var_length
    
    sequence = ''.join(sequence)

    return sequence

class TempDirectory(object):
    """ Temporary directory context manager """
    def __init__(self):
        self.name = os.path.abspath(str(uuid.uuid4()).replace('-', ''))
    def __enter__(self):
        os.mkdir(self.name)
        return self.name
    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.name)




'''
# General plot functions

'''

import matplotlib.pyplot as plt

import wgs_analysis.helpers as helpers


def plot_patient_legend(legend_filename):
	fig = plt.figure(figsize=(0.75, 2))
	ax = fig.add_subplot(111)
	ax.axis('off')

	artists = [plt.Circle((0, 0), color=c) for c in helpers.patient_cmap.values()]
	ax.legend(artists, helpers.patient_cmap.keys(), loc='upper right', title='Patient')

	fig.savefig(legend_filename, bbox_inches='tight')




'''
# Circos plotting

Python interface for circos plots of breakpoints, copy number histograms, and SNVs as stack boxes.  

'''

import collections
import os
import glob
import subprocess
import pandas as pd
import numpy as np


def create_circos_config(chromosomes, breakpoints, copynumbers, snvs, config_filename, svg_filename):

    template_filename = os.path.abspath(os.path.join(os.path.dirname(__file__), 'circos.conf'))

    def circos_chromosome(chromosome):
        return 'hs' + chromosome

    links_by_color = collections.defaultdict(list)
    breakpoints = breakpoints.drop_duplicates(['cluster_id', 'cluster_end'])
    for cluster_id, rows in breakpoints.groupby('cluster_id'):
        row1 = rows.iloc[0]
        row2 = rows.iloc[1]
        color = 'black'
        try:
            color = row1['link_color']
        except:
            pass
        link = ((str(cluster_id), circos_chromosome(row1['chrom']), str(row1['coord']), str(row1['coord']+1)),
                (str(cluster_id), circos_chromosome(row2['chrom']), str(row2['coord']), str(row2['coord']+1)))
        links_by_color[color].append(link)

    link_text = ''
    z = 20
    index = 0
    for color, links in sorted(links_by_color.items()):
        link_filename = config_filename + '.{0}.circos.link'.format(color)
        with open(link_filename, 'w') as link_file:
            for link in links:
                link_file.write('\t'.join(link[0]) + '\n')
                link_file.write('\t'.join(link[1]) + '\n')
        link_text += '\t<link {0}>\n'.format(color)
        link_text += '\t\tz={0}\n'.format(z)
        link_text += '\t\tthickness=2\n'
        link_text += '\t\tcolor={0}\n'.format(color)
        link_text += '\t\tfile={0}\n'.format(link_filename)
        link_text += '\t</link>\n'
        z += 1
        index += 1

    output_text = ''
    output_text += '\tdir={0}\n'.format(os.path.abspath(os.path.dirname(svg_filename)))
    output_text += '\tfile={0}\n'.format(os.path.basename(svg_filename))

    copies_text = ''
    track = 0.0
    copy_track_width = 0.04
    for id, copynumber in copynumbers.iteritems():
        copynumber['chrom'] = copynumber.apply(lambda row: circos_chromosome(row['chrom']), axis=1)
        colors = ['red', 'blue']
        hist_datas = [copynumber[copynumber['median'] > 0.0], copynumber[copynumber['median'] < 0.0]]
        for color, hist_data in zip(colors, hist_datas):
            hist_filename = config_filename + '.{0}.circos.cnv.{1}'.format(id, color)
            hist_data.to_csv(hist_filename, sep='\t', header=False, index=False, cols=['chrom', 'start', 'end', 'median'])
            copies_text += '<plot>\n'
            copies_text += 'type = histogram\n'
            copies_text += 'file = {0}\n'.format(hist_filename)
            copies_text += 'r0 = {0}r\n'.format(1.0 - copy_track_width * (track + 1.0) - 0.01)
            copies_text += 'r1 = {0}r\n'.format(1.0 - copy_track_width * track - 0.01)
            copies_text += 'min=-0.5\n'
            copies_text += 'max=0.5\n'
            copies_text += 'color = {0}\n'.format(color)
            copies_text += 'fill_under = yes\n'
            copies_text += 'fill_color = {0}\n'.format(color)
            copies_text += 'thickness = 0\n'
            copies_text += 'extend_bin = no\n'
            copies_text += 'background = no\n'
            copies_text += '</plot>\n'
        track += 1.0

    snv_tiles_filename = config_filename + '.snv.tiles'
    with open(snv_tiles_filename, 'w') as snv_tiles_file:
        for idx, row in snvs.iterrows():
            snv_tiles_file.write('\t'.join([circos_chromosome(row['chrom']), str(row['pos']), str(row['pos'])]) + '\n')

    snvs_track_width = 0.2
    snvs_text = ''
    snvs_text += 'type = tile\n'
    snvs_text += 'layers_overflow = grow\n'
    snvs_text += '<plot>\n'
    snvs_text += 'file = {0}\n'.format(snv_tiles_filename)
    snvs_text += 'r1 = {0}r\n'.format(1.0 - copy_track_width * track - 0.01)
    snvs_text += 'r0 = {0}r\n'.format(1.0 - copy_track_width * track - 0.01 - snvs_track_width)
    snvs_text += 'orientation = in\n'
    snvs_text += 'layers = 15\n'
    snvs_text += 'margin = 0.02u\n'
    snvs_text += 'thickness = 4\n'
    snvs_text += 'padding = 2\n'
    snvs_text += 'stroke_thickness = 4\n'
    snvs_text += 'stroke_color = orange\n'
    snvs_text += 'color = blue\n'
    snvs_text += '</plot>\n'

    with open(template_filename, 'r') as template:
        config_text = template.read()

    config_text = config_text.replace('!insert_svg_here!', output_text)
    config_text = config_text.replace('!insert_links_here!', link_text)
    config_text = config_text.replace('!insert_links_radius_here!', str(1.0 - copy_track_width * track - 0.01))
    config_text = config_text.replace('!insert_chromosomes_default_here!', 'no')
    config_text = config_text.replace('!insert_chromosomes_here!', ' '.join([circos_chromosome(a) for a in chromosomes]))
    config_text = config_text.replace('!insert_copies_here!', copies_text)
    config_text = config_text.replace('!insert_snvs_here!', snvs_text)

    with open(config_filename, 'w') as config_file:
        config_file.write(config_text)



def create_circos_plot(chromosomes, breakpoints, copynumbers, snvs, plot_filename):

    config_filename = plot_filename + '.tmp.conf'
    create_circos_config(chromosomes, breakpoints, copynumbers, snvs, config_filename, plot_filename)
    circos_bin = os.path.expanduser('~/Build/circos-0.52/bin/circos')
    subprocess.check_call('perl {0} -conf {1}'.format(circos_bin, config_filename), shell=True)
    for filename in glob.glob(os.path.abspath(plot_filename) + '.tmp*'):
        os.remove(filename)




import unittest
import numpy as np
import pandas as pd

import wgs_analysis.algorithms.rearrangement


def generate_random_breakpoints(num_breakpoints, max_position=10000, seed=123):
    np.random.seed(seed)

    chromosomes = ['1', '2']
    strands = ['+', '-']

    breakpoints = pd.DataFrame({
        'chromosome_1': np.random.choice(chromosomes, size=num_breakpoints),
        'chromosome_2': np.random.choice(chromosomes, size=num_breakpoints),
        'strand_1': np.random.choice(strands, size=num_breakpoints),
        'strand_2': np.random.choice(strands, size=num_breakpoints),
        'position_1': np.random.randint(max_position, size=num_breakpoints),
        'position_2': np.random.randint(max_position, size=num_breakpoints),
    })
    breakpoints['prediction_id'] = range(len(breakpoints.index))

    return breakpoints


def find_matches_brute(breakpoints_1, breakpoints_2, extend=500):
    matches = []

    for idx1, row1 in breakpoints_1.iterrows():
        for idx2, row2 in breakpoints_2.iterrows():

            match1 = (
                (row1['chromosome_1'] == row2['chromosome_1']) &
                (row1['strand_1'] == row2['strand_1']) &
                (abs(row1['position_1'] - row2['position_1']) <= extend) &
                (row1['chromosome_2'] == row2['chromosome_2']) &
                (row1['strand_2'] == row2['strand_2']) &
                (abs(row1['position_2'] - row2['position_2']) <= extend))
            match2 = (
                (row1['chromosome_1'] == row2['chromosome_2']) &
                (row1['strand_1'] == row2['strand_2']) &
                (abs(row1['position_1'] - row2['position_2']) <= extend) &
                (row1['chromosome_2'] == row2['chromosome_1']) &
                (row1['strand_2'] == row2['strand_1']) &
                (abs(row1['position_2'] - row2['position_1']) <= extend))

            if match1 or match2:
                matches.append({
                    'from_idx': row1['prediction_id'],
                    'to_idx': row2['prediction_id'],
                })

    matches = pd.DataFrame(matches)

    return matches


class BreakpointDatabaseTestCase(unittest.TestCase):
    def test_find_matching(self):
        extend = 500

        random_breakpoints = generate_random_breakpoints(1000, max_position=2000, seed=1)

        breakpoint_db = wgs_analysis.algorithms.rearrangement.BreakpointDatabase(random_breakpoints)

        test_breakpoints = generate_random_breakpoints(1000, max_position=2000, seed=2)

        matches = []

        for i, (idx, row) in enumerate(test_breakpoints.iterrows()):
            for match in breakpoint_db.query(row, extend=extend):
                matches.append({
                    'from_idx': match,
                    'to_idx': row['prediction_id'],
                })

        matches = pd.DataFrame(matches)

        test_matches = find_matches_brute(random_breakpoints, test_breakpoints, extend=extend)

        test_results = pd.merge(
            matches.assign(matches=1),
            test_matches.assign(test_matches=1),
            how='outer',
        )

        self.assertFalse(test_results['matches'].isnull().any())
        self.assertFalse(test_results['test_matches'].isnull().any())




import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn
import pylab
import scipy.cluster.hierarchy as sch
import scipy.spatial.distance as dst


def binary_heatmap(fig, df):

    seaborn.set_style('white')

    dat = df.values
    rownames = np.array(df.index.values)
    colnames = np.array(df.columns.values)

    # Compute and plot column dendrogram
    ax = fig.add_axes([0.0,0.8,1.0,0.2])
    D = dst.squareform(dst.pdist(dat.T, 'cityblock'))
    Y = sch.linkage(D, method='complete')
    Z = sch.dendrogram(Y, color_threshold=-1)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    # Plot distance matrix.
    ax = fig.add_axes([0.0,0.0,1.0,0.75])
    idx1 = np.lexsort(dat.T)
    idx2 = Z['leaves']
    dat = dat[idx1,:]
    dat = dat[:,idx2]
    rownames = rownames[idx1]
    colnames = colnames[idx2]
    im = ax.matshow(dat, aspect='auto', origin='lower', cmap=pylab.cm.Blues)
    ax.set_yticks([])
    ax.set_xticks(range(len(colnames)))
    ax.set_xticklabels(colnames)
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



def intensity_heatmap(fig, df, sort_rows=True):

    seaborn.set_style('white')

    dat = np.log(df.values + 1)
    rownames = np.array(df.index.values)
    colnames = np.array(df.columns.values)

    # Compute and plot column dendrogram
    ax = fig.add_axes([0.0,0.8,0.9,0.15])
    D = dst.squareform(dst.pdist(dat.T, 'cityblock'))
    Y = sch.linkage(D, method='complete')
    Z = sch.dendrogram(Y, color_threshold=-1)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    # Plot distance matrix.
    ax = fig.add_axes([0.0,0.0,0.9,0.75])
    D2 = dst.squareform(dst.pdist(dat, 'cityblock'))
    Y2 = sch.linkage(D2, method='complete')
    Z2 = sch.dendrogram(Y2, color_threshold=-1, no_plot=True)
    idx1 = Z2['leaves']
    idx2 = Z['leaves']
    if sort_rows:
        dat = dat[idx1,:]
        rownames = rownames[idx1]
    dat = dat[:,idx2]
    colnames = colnames[idx2]
    im = ax.matshow(dat, aspect='auto', origin='lower', cmap=pylab.cm.Blues)
    ax.set_yticks([])
    ax.set_xticks(range(len(colnames)))
    ax.set_xticklabels(colnames)
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.grid(True, which='minor', color='black')
    
    # Plot colorbar.
    axcolor = fig.add_axes([0.91,0.1,0.02,0.6])
    cb = pylab.colorbar(im, cax=axcolor, drawedges=True)
    cb.outline.set_linewidth(0)
    cb.dividers.set_linewidth(0)
    axcolor.set_ylabel('log(read count + 1)')

    return Z




import sys
import csv
import subprocess
import math
import argparse
import os
import itertools
import string
import gzip
import pandas as pd
import numpy as np
from scipy import stats


###
# Annnotation and filtering parameters
###

align_prob_score_min = 0.9
valid_prob_score_min = 0.5
chimeric_prob_score_min = 0.9

high_conf_min_tumour_count_max = 4
high_conf_min_num_split = 2

max_other_libraries = 0



###
# Helper functions
###

def get_counts_columns(breakpoints):

    return filter(lambda a: a.endswith('_count'), breakpoints.columns)


def shortened_site_names(row):
    shortened = list()
    for site_id in sorted(row.loc[row['num_reads'] > 0, 'site_id'].unique()):
        shortened.append(''.join(a[0] for a in site_id.split('_')))
    return '_'.join(shortened)


def calculate_site_names(breakpoint_site):
    """ Calculate unique identifier to each presence/absence group
    """
    return breakpoint_site.set_index(['prediction_id'])\
                          .groupby(level=0).apply(shortened_site_names)\
                          .reset_index()\
                          .rename(columns={0:'site_names'})


def annotate_site_names(breakpoint, breakpoint_site):
    """ Add `site_names` column based on presence absence of breakpoints
    """

    breakpoint_site_names = calculate_site_names(breakpoint_site)

    breakpoint = breakpoint.merge(breakpoint_site_names)

    return breakpoint


def annotate_tumour_count_max(breakpoints):
    """ Annotation of max tumour count across tumour samples
    """

    counts_columns = get_counts_columns(breakpoints)

    # Max tumour count across sites
    breakpoints.set_index('prediction_id', inplace=True)
    breakpoints['tumour_count_max'] = breakpoints[counts_columns].max(axis='columns')
    breakpoints.reset_index(inplace=True)

    # Abbreviated site name group
    breakpoints = annotate_site_names(breakpoints)

    return breakpoints


def filter_breakpoints(breakpoints):
    """ Filtering germline and putative false positives
    """

    # No germline
    breakpoints = breakpoints[breakpoints['normal_blood_count'] == 0]

    # No common rearrangements
    breakpoints = breakpoints[breakpoints['common'] <= max_other_libraries]

    # Breakpoint probability filter
    breakpoints = breakpoints[breakpoints['align_prob'] > align_prob_score_min]
    breakpoints = breakpoints[breakpoints['valid_prob'] > valid_prob_score_min]
    breakpoints = breakpoints[breakpoints['chimeric_prob'] > chimeric_prob_score_min]

    # Filter small deletions
    breakpoints = breakpoints[(breakpoints['type'] != 'deletion') | (breakpoints['break_distance'] > 1000)]

    return breakpoints


def stack_counts(breakpoints):
    """ Convert to stacked counts table
    """


    counts_columns = get_counts_columns(breakpoints)
    break_counts = breakpoints.set_index('prediction_id')[counts_columns].stack().reset_index()
    break_counts.columns = ['prediction_id', 'site_id', 'num_span']
    break_counts['site_id'] = break_counts['site_id'].apply(lambda a: a[:-len('_count')])
    breakpoints = breakpoints.drop(counts_columns + ['normal_blood_count'], axis=1)
    breakpoints = breakpoints.merge(break_counts, left_on='prediction_id', right_on='prediction_id')

    return breakpoints


def stack_sides(breakpoints):
    """ Convert to stacked cluster end 1 and 2 table
    """

    breakpoints_common = breakpoints.filter(regex='[^12]$')

    def remove_side_suffix(column):
        if column.endswith('_1') or column.endswith('_2'):
            return column[:-2]
        elif column.endswith('1') or column.endswith('2'):
            return column[:-1]
        return column

    breakpoints_side_1 = breakpoints.set_index('prediction_id').filter(regex='1$', axis=1).rename(columns=remove_side_suffix).reset_index().drop_duplicates()
    breakpoints_side_1['cluster_end'] = 0
    
    breakpoints_side_2 = breakpoints.set_index('prediction_id').filter(regex='2$', axis=1).rename(columns=remove_side_suffix).reset_index().drop_duplicates()
    breakpoints_side_2['cluster_end'] = 1

    breakpoints_sides = pd.concat([breakpoints_side_1, breakpoints_side_2], ignore_index=True)

    breakpoints = breakpoints_common.merge(breakpoints_sides, left_on='prediction_id', right_on='prediction_id')

    breakpoints = breakpoints.set_index(['prediction_id', 'cluster_end']).reset_index()

    return breakpoints


def annotate_high_confidence(breakpoints):
    """ Annotation of high confident breakpoints
    """

    # High confidence breakpoints
    breakpoints['high_conf'] = ((breakpoints['tumour_count_max'] >= high_conf_min_tumour_count_max) &
                                (breakpoints['num_split'] >= high_conf_min_num_split)) * 1

    return breakpoints



def add_annotations(breakpoints):

    breakpoints = annotate_tumour_count_max(breakpoints)
    breakpoints = annotate_site_names(breakpoints)

    breakpoints = filter_breakpoints(breakpoints)

    breakpoints = stack_counts(breakpoints)
    breakpoints = stack_sides(breakpoints)

    breakpoints = annotate_high_confidence(breakpoints)

    breakpoints = breakpoints.sort(['prediction_id', 'site_id'])

    return breakpoints


def annotate_expression_correlation(breakpoints, expression):
    """ Add presence correlation with expression expr_corr column
    """

    breakpoints = breakpoints.merge(expression)

    breakpoints.set_index(['prediction_id', 'cluster_end'], inplace=True)

    breakpoints['presences'] = (breakpoints['num_span'] > 0) * 1
    
    expr_corr = breakpoints[['presences', 'recentred_expression']].groupby(level=[0, 1]).corr().fillna(0.0)
    expr_corr = expr_corr.xs('presences', level=2)['recentred_expression']
    
    breakpoints['expr_corr'] = expr_corr

    breakpoints.reset_index(inplace=True)
    
    breakpoints = breakpoints.drop('presences', axis=1)

    return breakpoints


def annotate_size_class(data):
    """ Add the categorical column size_class.

    Expects columns:
        - position_1
        - position_2
        - type
    """

    length = (data['position_1'] - data['position_2']).abs()
    data['size_class'] = pd.cut(
        length,
        [0, 1e4, 1e6, 1e8, 1e10],
        labels=['0-1K', '1K-1M', '1-100M', '>100M'])
    data['size_class'] = data['size_class'].astype(str)
    data.loc[data['type'] == 'translocation', 'size_class'] = 'Tr'
    data['size_class'] = pd.Categorical(
        data['size_class'],
        categories=reversed(['0-1K', '1K-1M', '1-100M', '>100M', 'Tr']))

    return data





"""
SNV annotation functionality

Annotations for SNV data.

Functions take pandas tables and add additional columns.
"""

import itertools
import pandas as pd

def annotate_cluster(snvs, bandwidth=5000, min_spacing=4):
    """ Add cluster_id and cluster_size columnns for clusters of snvs
    within bandwidth nt of each other.  cluster_size counts only those snvs
    that are at least min_spacing apart
    """
    class PositionClusterKey(object):
        def __init__(self, bandwidth):
            self.bandwidth = bandwidth
            self.start = None
            self.previous = None
        def __call__(self, position):
            if self.start is None or position - self.previous > self.bandwidth:
                self.start = position
            self.previous = position
            return self.start
    cluster_id = 0
    clusters_table = list()
    for chromosome, chr_group in snvs[['chrom', 'coord']].drop_duplicates().groupby('chrom'):
        for start, pos_group in itertools.groupby(sorted(chr_group['coord']), key=PositionClusterKey(bandwidth)):
            pos_group = list(pos_group)
            cluster_size = 0
            previous_position = None
            for position in sorted(pos_group):
                if previous_position is None or position - previous_position >= min_spacing:
                    cluster_size += 1
                previous_position = position
            for position in pos_group:
                clusters_table.append((chromosome, position, cluster_id, cluster_size))
            cluster_id += 1
    clusters_table = pd.DataFrame(clusters_table, columns=['chrom', 'coord', 'cluster_id', 'cluster_size'])
    snvs = snvs.merge(clusters_table, left_on=['chrom', 'coord'], right_on=['chrom', 'coord'])
    return snvs


def annotate_kataegis(snvs, min_cluster_size=5):
    """ Annotate kataegis_id column, identifier for the kataegis event the snv
    is putatively part of
    """

    snvs = annotate_cluster(snvs)

    kataegis_id = 0
    snvs['kataegis_id'] = None
    for cluster_id, group in snvs.groupby('cluster_id'):
        if group['cluster_size'].iloc[0] < min_cluster_size:
            continue
        for idx, row in group.iterrows():
            snvs['kataegis_id'][idx] = kataegis_id
        kataegis_id += 1
    return snvs


def shortened_site_names(row):
    shortened = list()
    for site_id in sorted(row.loc[row['alt_counts'] > 0, 'site_id'].unique()):
        shortened.append(''.join(a[0] for a in site_id.split('_')))
    return '_'.join(shortened)


def annotate_site_names(snvs):
    """ Annotate unique identifier to each presence/absence group
    """
    snvs.set_index(['chrom', 'coord'], inplace=True)
    snvs['site_names'] = snvs.groupby(level=[0, 1]).apply(shortened_site_names)
    snvs.reset_index(inplace=True)

    return snvs




'''
# Single cell composite plots

'''

import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.patches import Rectangle as Rectangle
from mpl_toolkits.axes_grid import Divider, LocatableAxes, Size
import seaborn

import wgs_analysis.helpers as helpers
import wgs_analysis.plots as plots
import wgs_analysis.plots.colors
import wgs_analysis.plots.utils
import wgs_analysis.plots.trees


def plot_tree_genotype_samples(tree, genotype_table, sample_table, prevalence_type='count', event_colors=None, small_values=None):
    """ Plot a genotype tree with sample specific genotype prevalences/counts

    Args:
        tree (dollo.trees.TreeNode): tree to plot
        genotype_table (pandas.DataFrame): table of genotypes per node
        sample_table (pandas.DataFrame): table of sample prevalences/counts

    KwArgs:
        prevalence_type (str): type of barplot, 'count' or 'proportion'
        event_colors (dict): event colors keyed by event id
        small_values (float): label positive values less than or equal this value

    The genotype table should have node labels as row indices and event ids (ids of 
    snv/breakpoint clusters) as column labels.  Each 0/1 entry represents the
    presence or absence of the event for that node.  The sample table should have
    sample ids as row indices and leaf names as column labels.

    """

    leaf_names = [str(leaf.name) for leaf in tree.leaves]
    events = [str(event) for event in genotype_table.columns]

    num_samples = len(sample_table.index)
    num_events = len(genotype_table.columns)

    if event_colors is None:
        if num_events > 8:
            palette = seaborn.color_palette('husl', num_events)
        else:
            palette = seaborn.color_palette('Set2', num_events)
        event_colors = dict(zip(genotype_table.columns.values, palette))

    fig = plt.figure(figsize=(20, 3))
    plots.utils.setup_plot()

    # Plot grid
    gs = gridspec.GridSpec(1, num_samples + 2, height_ratios=[1], width_ratios=[2,2] + [1] * num_samples)

    # Plot tree
    tree_ax = plt.subplot(gs[0])

    plots.trees.plot_tree(tree_ax, tree, extend_leaf_branch=True, landscape=True, flip=False)
    plots.trees.plot_branch_events(tree_ax, tree, genotype_table, event_colors=event_colors)
    plots.utils.shift_lims(tree_ax, 12, 12)

    tree_ax.set_yticks(())
    tree_ax.yaxis.set_tick_params(right='off')
    tree_ax.spines['right'].set_visible(False)
    tree_ax.set_title('Clone Phylogeny')

    # Genotype plots
    genotype_ax = plots.utils.setup_axes(plt.subplot(gs[1]))

    width = 1.0
    height = 0.8

    for leaf in tree.leaves:

        offset_x = 0
        offset_y = -height * 0.5

        for idx, (event_id, present) in enumerate(genotype_table.loc[leaf.label].iteritems()):

            widget_x = width * idx - offset_x
            widget_y = leaf.y + offset_y

            color = 'white'
            if present == 1:
                color = event_colors[event_id]

            genotype_ax.add_patch(Rectangle((widget_x, widget_y), width, height,
                facecolor=color, edgecolor='.75', linewidth=0.25, zorder=10))

        genotype_ax.add_patch(Rectangle((0, leaf.y + offset_y), num_events, height,
            facecolor=(0,0,0,0), edgecolor='k', linewidth=1., zorder=10))

    genotype_ax.set_xticks(np.arange(num_events) + 0.5)
    genotype_ax.set_xticklabels(events)
    genotype_ax.set_yticks(xrange(len(leaf_names)))
    genotype_ax.set_yticklabels(leaf_names, ha='left')
    genotype_ax.set_xlim((0, num_events))
    genotype_ax.xaxis.set_tick_params(which='both', top=False, bottom=False, labeltop=False, labelbottom=True, labelsize=6, pad=0)
    genotype_ax.yaxis.set_tick_params(which='both', right=False, left=False, labelright=False, labelleft=True, labelsize=10, pad=0)
    genotype_ax.spines['left'].set_visible(False)
    genotype_ax.spines['right'].set_visible(False)
    genotype_ax.spines['top'].set_visible(False)
    genotype_ax.spines['bottom'].set_visible(False)
    genotype_ax.grid(False)
    genotype_ax.set_title('Observed Genotypes')

    # Plot barplots
    max_cell_count = sample_table.max().max()

    for idx, (sample_id, row) in enumerate(sample_table.iterrows()):

        ax = plots.utils.setup_axes(plt.subplot(gs[idx+2]))

        row = row.reindex(index=leaf_names)

        ind = np.arange(len(row))

        ax.barh(ind, row.values, align='center', height=1.0, color='0.8')

        if small_values is not None:
            for i, v in zip(ind, row.values):
                if v <= small_values and v > 0:
                    ax.text(v+3, i, '{0}'.format(int(v)), fontsize=8, ha='center', va='center')

        ax.grid(False)
        ax.spines['bottom'].set_position(('outward', 0))
        ax.spines['left'].set_position(('outward', 0))

        if prevalence_type == 'count':
            ax.set_xlim((ax.get_xlim()[0], max_cell_count * 1.05))
        elif prevalence_type == 'proportion':
            ax.set_xlim((ax.get_xlim()[0], 1.0))
        ax.set_ylim((min(ind) - 0.5, max(ind) + 0.5))
        ax.yaxis.set_tick_params(left='off', labelleft='off')
        ax.set_title(sample_id)
        ax.set_ylabel('')

        if idx == len(sample_table.index) - 1:
            ax.set_xlabel(prevalence_type)
            if prevalence_type == 'count':
                ax.set_xticks(np.linspace(0, max_cell_count, 3))
                for tick in ax.xaxis.get_major_ticks():
                    tick.label.set_fontsize(8) 
            elif prevalence_type == 'proportion':
                ax.set_xticks((0., 0.5, 1.0))
        else:
            ax.set_xticks(np.linspace(0, max_cell_count, 3))
            ax.set_xticklabels([])
            ax.set_xlabel('')

    tree_ax.set_ylim(ax.get_ylim())
    genotype_ax.set_ylim(tree_ax.get_ylim())

    fig.subplots_adjust(wspace=0.1, hspace=0.1, bottom=0.0)

    return fig



def plot_tree_genotype(tree, genotype_table, event_colors=None, palette=None):
    """ Plot a genotype tree with sample specific genotype prevalences/counts

    Args:
        tree (dollo.trees.TreeNode): tree to plot
        genotype_table (pandas.DataFrame): table of genotypes per node

    KwArgs:
        event_colors (dict): event colors keyed by event id
        use_same_colors (dict): color palette to use. If None will automatically choose based on number of genotypes. 

    The genotype table should have node labels as row indices and event ids (ids of 
    snv/breakpoint clusters) as column labels.  Each 0/1 entry represents the
    presence or absence of the event for that node.

    """

    leaf_names = [str(leaf.name) for leaf in tree.leaves]
    events = [str(event) for event in genotype_table.columns]

    num_leaves = len(leaf_names)
    num_events = len(genotype_table.columns)

    if event_colors is None:
        if palette is not None:
            palette = seaborn.color_palette(palette, num_events)
        
        else:
            if num_events > 8:
                palette = seaborn.color_palette('husl', num_events)
            else:
                palette = seaborn.color_palette('Set2', num_events)

        event_colors = dict(zip(genotype_table.columns.values, palette))

    fig = plt.figure()
    plots.utils.setup_plot()

    scale = 0.4

    horiz = [
        scale * num_leaves,   # tree
        0.2,                  # spacer
        scale * num_events,   # genotype
    ]

    vert = [
        scale * num_leaves,   # genotype/tree
    ]

    horiz = [Size.Fixed(a) for a in horiz]
    vert = [Size.Fixed(a) for a in vert]
    rect = (0.1, 0.1, 0.8, 0.8)

    tree_ax = fig.add_axes(rect, label='tree')
    genotype_ax = fig.add_axes(rect, label='genotype')

    divider = Divider(fig, rect, horiz, vert, aspect=False)

    tree_ax.set_axes_locator(divider.new_locator(nx=0, ny=0))
    genotype_ax.set_axes_locator(divider.new_locator(nx=2, ny=0))

    plots.utils.setup_axes(tree_ax)
    plots.utils.setup_axes(genotype_ax)

    # Plot tree

    plots.trees.plot_tree(tree_ax, tree, extend_leaf_branch=True, landscape=True, flip=False)
    plots.trees.plot_branch_events(tree_ax, tree, genotype_table, event_colors=event_colors)
    plots.utils.shift_lims(tree_ax, 12, 12)

    tree_ax.set_yticks(())
    tree_ax.yaxis.set_tick_params(right='off')
    tree_ax.spines['right'].set_visible(False)
    tree_ax.set_title('Phylogeny', y=0.9)

    width = 1.0
    height = 0.8

    for leaf in tree.leaves:

        offset_x = 0
        offset_y = -height * 0.5

        for idx, (event_id, present) in enumerate(genotype_table.loc[leaf.label].iteritems()):

            widget_x = width * idx - offset_x
            widget_y = leaf.y + offset_y

            color = 'white'
            if present == 1:
                color = event_colors[event_id]

            genotype_ax.add_patch(Rectangle((widget_x, widget_y), width, height,
                facecolor=color, edgecolor='.75', linewidth=0.25, zorder=10))

        genotype_ax.add_patch(Rectangle((0, leaf.y + offset_y), num_events, height,
            facecolor=(0,0,0,0), edgecolor='k', linewidth=1., zorder=10))

    genotype_ax.set_xticks(np.arange(num_events) + 0.5)
    genotype_ax.set_xticklabels(events)
    genotype_ax.set_xlabel('PyClone cluster')
    genotype_ax.set_yticks(xrange(len(leaf_names)))
    genotype_ax.set_yticklabels(leaf_names, ha='left')
    genotype_ax.set_xlim((0, num_events))
    genotype_ax.xaxis.set_tick_params(which='both', top=False, bottom=False, labeltop=False, labelbottom=True, labelsize=10, pad=-10)
    genotype_ax.yaxis.set_tick_params(which='both', right=False, left=False, labelright=False, labelleft=True, labelsize=10, pad=0)
    genotype_ax.spines['left'].set_visible(False)
    genotype_ax.spines['right'].set_visible(False)
    genotype_ax.spines['top'].set_visible(False)
    genotype_ax.spines['bottom'].set_visible(False)
    genotype_ax.grid(False)
    genotype_ax.set_title('Genotypes', y=0.9)

    tree_ax.set_ylim((-0.5, float(len(leaf_names)) + 0.5))
    genotype_ax.set_ylim((-0.5, float(len(leaf_names)) + 0.5))

    return fig






import os
import seaborn
import lda
import scipy.stats
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pyfaidx
import matplotlib.font_manager
import pkg_resources


def reverse_complement(sequence):
    return sequence[::-1].translate(str.maketrans('ACTGactg','TGACtgac'))


def complement(sequence):
    return sequence.translate(str.maketrans('ACTGactg','TGACtgac'))


def fit_sample_signatures(snvs_table, sig_prob, subset_col):

    # Filter samples with fewer than 100 SNVs
    snv_counts = snvs_table.groupby(subset_col).size()
    snvs_table = snvs_table.set_index(subset_col).loc[snv_counts[snv_counts > 100].index].reset_index()

    tri_nuc_table = (snvs_table.groupby([subset_col, 'tri_nuc_idx'])
        .size().unstack().fillna(0).astype(int))

    if len(tri_nuc_table.index) == 0:
        return pd.DataFrame()

    model = lda.LDA(n_topics=len(sig_prob.columns), random_state=0, n_iter=10000, alpha=0.01)
    model.components_ = sig_prob.values.T

    sample_sig = model.transform(tri_nuc_table.values, max_iter=1000)

    sample_sig = pd.DataFrame(sample_sig, index=tri_nuc_table.index, columns=sig_prob.columns)
    sample_sig.index.name = 'Sample'
    sample_sig.columns = [a[len('Signature '):] for a in sample_sig.columns]
    sample_sig.columns.name = 'Signature'

    return sample_sig


def plot_signature_heatmap(sample_sig):
    if sample_sig.shape[0] <= 1:
        return plt.figure(figsize=(8,5))
    g = seaborn.clustermap(sample_sig, figsize=(8,5))
    g.ax_heatmap.set_yticklabels(g.ax_heatmap.get_yticklabels(), rotation=0)
    return g.fig


def test_ancestral_descendant(data):
    data = dict(list(data.groupby(by=lambda a: a.endswith('Node0'))))
    return scipy.stats.mannwhitneyu(data[True], data[False])[1]


def plot_signature_boxplots(sample_sig, pvalue_threshold=0.01):

    # Show only signatures with p-value less than specified
    test_pvalue = sample_sig.apply(test_ancestral_descendant)
    data = sample_sig.loc[:,test_pvalue[test_pvalue < pvalue_threshold].index]

    data = data.stack()
    data.name = 'Proportion'
    data = data.reset_index()
    data['is_node0'] = data['Sample'].apply(lambda a: a.endswith('Node0'))
    data['Branch'] = data['is_node0'].apply(lambda a: ('Descendant', 'Ancestral')[a])

    sig_order = np.sort(data['Signature'].unique().astype(int)).astype(str)

    g = seaborn.FacetGrid(data, col='Signature', col_order=sig_order, col_wrap=5, margin_titles=True, sharey=False)
    g.map_dataframe(seaborn.boxplot, x='Branch', y='Proportion', fliersize=0., color='0.75')
    g.map_dataframe(seaborn.stripplot, x='Branch', y='Proportion', jitter=True, color='k',
        linewidth=0, split=False)

    for signature, ax in zip(g.col_names, g.axes):
        ax.set_title('Signature {0}\n  (p = {1:.1e})'.format(signature, test_pvalue.loc[signature]))
        yticks = ax.get_yticks()
        ax.set_yticks(yticks[yticks >= 0.])

    new_xticklabels = list()
    for label in (a.get_text() for a in g.axes[0].get_xticklabels()):
        n = len(data.loc[data['Branch'] == label, 'Sample'].unique())
        label = '{}\nn={}'.format(label, n)
        new_xticklabels.append(label)
    g.axes[0].set_xticklabels(new_xticklabels)

    seaborn.despine(offset=10, trim=True)

    plt.tight_layout()

    return g.fig


def load_signature_probabilities():
    """ Load a dataframe of cosmic signature probabilities.
    """
    sig_prob_filename = pkg_resources.resource_filename('wgs_analysis', 'data/signatures_probabilities.tsv')

    sig_prob = pd.read_csv(sig_prob_filename, sep='\t')

    sig_prob['tri_nucleotide_context'] = sig_prob['Trinucleotide']
    sig_prob['tri_nuc_idx'] = range(len(sig_prob.index))
    sig_prob['ref'] = sig_prob['Substitution Type'].apply(lambda a: a.split('>')[0])
    sig_prob['alt'] = sig_prob['Substitution Type'].apply(lambda a: a.split('>')[1])

    # Original
    sig1 = sig_prob[['tri_nuc_idx', 'ref', 'alt', 'tri_nucleotide_context']].copy()

    # Reverse complement
    sig2 = sig_prob[['tri_nuc_idx', 'ref', 'alt', 'tri_nucleotide_context']].copy()
    for col in ['ref', 'alt', 'tri_nucleotide_context']:
        sig2[col] = sig2[col].apply(reverse_complement)

    # Signatures in terms of ref and alt
    sigs = pd.concat([sig1, sig2], ignore_index=True)

    # Probability matrix
    signature_cols = filter(lambda a: a.startswith('Signature'), sig_prob.columns)
    sig_prob = sig_prob.set_index('tri_nuc_idx')[signature_cols]

    return sigs, sig_prob


def normalize_trinucleotides(data):
    """ Normalize trinucleotide reverse complementing consistent with cosmic

    Args:
        data (DataFrame): input snv data
    
    Returns:
        DataFrame: snv data with output columns added
        DataFrame: snv type, trinucleotide, ordered for plotting

    Required columns:
        - ref
        - alt
        - tri_nucleotide_context

    Output columns:
        - mutation_type
        - norm_mutation_type
        - tri_nucleotide_context
        - norm_tri_nucleotide_context
    """

    normalized_mutation_types = ['C>A', 'C>G', 'C>T', 'T>A', 'T>C', 'T>G']

    norm_type_map = []

    for ref in ['A', 'C', 'T', 'G']:
        for alt in ['A', 'C', 'T', 'G']:
            if ref == alt:
                continue

            mutation_type = ref + '>' + alt
            is_revcomp = mutation_type not in normalized_mutation_types

            for left in ['A', 'C', 'T', 'G']:
                for right in ['A', 'C', 'T', 'G']:

                    trinuc = left + ref + right

                    norm_mutation_type = mutation_type
                    norm_trinuc = trinuc

                    if mutation_type not in normalized_mutation_types:
                        norm_mutation_type = complement(norm_mutation_type)
                        norm_trinuc = reverse_complement(norm_trinuc)

                    norm_type_map.append({
                        'ref': ref,
                        'alt': alt,
                        'mutation_type': mutation_type,
                        'norm_mutation_type': norm_mutation_type,
                        'tri_nucleotide_context': trinuc,
                        'norm_tri_nucleotide_context': norm_trinuc,
                    })

    norm_type_map = pd.DataFrame(norm_type_map)

    ordered_types = (
        norm_type_map[['norm_mutation_type', 'norm_tri_nucleotide_context']]
        .drop_duplicates()
        .sort_values(by=['norm_mutation_type', 'norm_tri_nucleotide_context']))

    data = data.merge(norm_type_map, on=['ref', 'alt', 'tri_nucleotide_context'], how='left')

    mismerged = data[data['norm_mutation_type'].isnull()]
    if not mismerged.empty:
        raise Exception(f'unable to normalize, example: {mismerged.iloc[0]}')

    assert not data['norm_mutation_type'].isnull().any()
    assert not data['norm_tri_nucleotide_context'].isnull().any()

    return data, ordered_types


mut_type_colors = {
    'C>A': [3/256,189/256,239/256],
    'C>G': [1/256,1/256,1/256],
    'C>T': [228/256,41/256,38/256],
    'T>A': [203/256,202/256,202/256],
    'T>C': [162/256,207/256,99/256],
    'T>G': [236/256,199/256,197/256],
}


def plot_mutation_spectra(data, count_column=None, fig=None):
    """ Plot 96 channel mutation spectra.

    Args:
        data (DataFrame): input snv data

    KwArgs:
        count_column (str): column designated already tabulated counts/proportions
    
    Returns:
        Figure: mutation spectra figure

    Required columns:
        - ref
        - alt
        - tri_nucleotide_context
    """

    data, ordered_types = normalize_trinucleotides(data)

    if count_column is None:
        plot_data = (
            data.groupby(['norm_mutation_type', 'norm_tri_nucleotide_context'])
            .size().rename('count'))
        count_column = 'count'

    else:
        plot_data = (
            data.set_index(['norm_mutation_type', 'norm_tri_nucleotide_context']))

    plot_data = plot_data.reindex(
        index=pd.MultiIndex.from_frame(ordered_types)).fillna(0).reset_index()

    plot_data['index'] = range(plot_data.shape[0])

    font = matplotlib.font_manager.FontProperties()
    font.set_family('monospace')

    if fig is None:
        fig = plt.figure(figsize=(16, 3), dpi=300)

    for mut_type, mut_type_data in plot_data.groupby('norm_mutation_type'):
        plt.bar(x='index', height=count_column, data=mut_type_data, label=mut_type, color=mut_type_colors[mut_type])
    plt.xticks(plot_data['index'], plot_data['norm_tri_nucleotide_context'], rotation=90, fontproperties=font)
    plt.xlim((-1, 97))
    plt.legend(bbox_to_anchor=(1, 1), loc='upper left')
    seaborn.despine(trim=True)

    return fig


def plot_cohort_mutation_signatures(
    sig_prob_filename,
    snvs_table,
    snv_nodes_table,
):
    """ Plot cohort wide clone specific mutation signatures.

    Args:
        sig_prob_filename (str): cosmic signature probability matrix
        snvs_table (pandas.DataFrame): table of per snv information including Trinucleotide
        snv_nodes_table (pandas.DataFrame): table of per snv per clone information

    """
    sigs, sig_prob = load_signature_probabilities()

    results = {}

    #
    # Per sample signatures
    #

    snvs_table = snvs_table[snvs_table['tri_nucleotide_context'].notnull()]
    snvs_table = snvs_table.merge(sigs)

    # Simple filter for variant sample presence
    snvs_table = snvs_table[snvs_table['alt_counts'] > 0]

    snvs_table['patient_sample_id'] = snvs_table['patient_id'] + '_' + snvs_table['sample_id']
    sample_sig = fit_sample_signatures(snvs_table, sig_prob, 'patient_sample_id')

    results['samples_table'] = sample_sig.copy()
    results['samples_heatmap'] = plot_signature_heatmap(sample_sig)

    #
    # Per node signatures
    #

    # Tri nucleotides from snvs table
    cohort_tri_nuc = snvs_table[['chrom', 'coord', 'ref', 'alt', 'tri_nuc_idx']].drop_duplicates()
    snv_nodes_table = snv_nodes_table.merge(cohort_tri_nuc)

    snv_nodes_table['patient_node_id'] = snv_nodes_table['patient_id'] + '_Node' + snv_nodes_table['node'].astype(str)
    node_sig = fit_sample_signatures(snv_nodes_table, sig_prob, 'patient_node_id')

    results['node_table'] = node_sig.copy()
    results['node_heatmap'] = plot_signature_heatmap(node_sig)
    results['node_signature_boxplots'] = plot_signature_boxplots(node_sig)

    return results


def calculate_tri_nucleotide_context(data, ref_genome_fasta):
    """ Calculate trinucleotide context and add as column
    """

    ref_genome = pyfaidx.Fasta(ref_genome_fasta, rebuild=False)

    data['tri_nucleotide_context'] = None

    for idx, row in data.iterrows():
        tnc = str(ref_genome[row['chrom']][row['coord']-2:row['coord']+1])
        assert len(tnc) == 3
        assert tnc[1] == row['ref']
        data.loc[idx, 'tri_nucleotide_context'] = tnc

    assert data['tri_nucleotide_context'].notnull().all()

    return data





'''
# Clonal composition plots

'''

import matplotlib.pyplot as plt

import wgs_analysis.helpers as helpers


def plot_mixture_class_legend(legend_filename):
	def create_marker_artist(m):
	    return plt.Line2D(range(1), range(1), color='w', marker=m, markerfacecolor='k')

	fig = plt.figure(figsize=(0.75, 2))
	ax = fig.add_subplot(111)
	ax.axis('off')

	artists = [create_marker_artist(m) for m in helpers.phylogenetic_class_markers.values()]
	labels = [helpers.phylogenetic_class_labels[m] for m in helpers.phylogenetic_class_markers.keys()]
	ax.legend(artists, labels, loc='lower right', title='Mixture\nclassification')

	fig.savefig(legend_filename, bbox_inches='tight')




import pandas as pd
import numpy as np


def position_segment_merge(positions, segments):
    """
    Merge positions with segments that contain them
    
    Args:
        positions (pandas.DataFrame): ['chrom', 'coord'] columns required
        segments (pandas.DataFrame): ['chrom', 'start', 'end'] columns required

    Returns:
        pandas.DataFrame: merged table with ['chrom', 'coord', 'start', 'end'] columns

    Assuming a set of non-overlapping segments, merge a set of positions so that
    each entry in the new table provides coord and containing segment start/end
    """

    # Check if any segments overlap
    for chromosome, df in segments.groupby('chrom'):
        if (df['start'].values[1:] < df['end'][:-1]).any():
            raise Exception('overlapping segments')

    positions = positions[['chrom', 'coord']].copy()
    segments = segments[['chrom', 'start', 'end']].copy()
    
    segments['chrom_1'] = segments['chrom']
    segments['coord'] = segments['start']
    
    positions['is_pos'] = 1
    segments['is_pos'] = 0
    
    merged = pd.concat([positions, segments], ignore_index=True)\
               .sort_values(['chrom', 'coord', 'is_pos'])\
               .drop_duplicates()
    
    merged['chrom_1'] = merged['chrom_1'].fillna(method='ffill')
    merged['start'] = merged['start'].fillna(method='ffill')
    merged['end'] = merged['end'].fillna(method='ffill')
    
    merged = merged[merged['is_pos'] == 1]
    
    merged = merged[(merged['chrom'] == merged['chrom_1']) &
                    (merged['coord'] >= merged['start']) &
                    (merged['coord'] <= merged['end'])]

    merged = merged.drop(['is_pos', 'chrom_1'], axis=1)
    merged['start'] = merged['start'].astype(int)
    merged['end'] = merged['end'].astype(int)

    return merged


def vrange(starts, lengths):
    """ Create concatenated ranges of integers for multiple start/length

    Args:
        starts (numpy.array): starts for each range
        lengths (numpy.array): lengths for each range (same length as starts)

    Returns:
        numpy.array: concatenated ranges

    See the following illustrative example:

        starts = np.array([1, 3, 4, 6])
        lengths = np.array([0, 2, 3, 0])

        print vrange(starts, lengths)
        >>> [3 4 4 5 6]

    """
    
    # Repeat start position index length times and concatenate
    cat_start = np.repeat(starts, lengths)

    # Create group counter that resets for each start/length
    cat_counter = np.arange(lengths.sum()) - np.repeat(lengths.cumsum() - lengths, lengths)

    # Add group counter to group specific starts
    cat_range = cat_start + cat_counter

    return cat_range


def interval_position_overlap(intervals, positions):
    """ Map intervals to contained positions

    Args:
        intervals (numpy.array): start and end of intervals with shape (N,2) for N intervals
        positions (numpy.array): positions, length M, must be sorted

    Returns:
        numpy.array: interval index, length L (arbitrary)
        numpy.array: position index, length L (same as interval index)

    Given a set of possibly overlapping intervals, create a mapping of positions that are contained
    within those intervals.

    """

    # Search for start and end of each interval in list of positions
    start_pos_idx = np.searchsorted(positions, intervals[:,0])
    end_pos_idx = np.searchsorted(positions, intervals[:,1])

    # Calculate number of positions for each segment
    lengths = end_pos_idx - start_pos_idx

    # Interval index for mapping
    interval_idx = np.repeat(np.arange(len(lengths)), lengths)

    # Position index for mapping 
    position_idx = vrange(start_pos_idx, lengths)

    return interval_idx, position_idx


def interval_position_overlap_unsorted(intervals, positions):
    """ Map intervals to contained positions

    Args:
        intervals (numpy.array): start and end of intervals with shape (N,2) for N intervals
        positions (numpy.array): positions, length M

    Returns:
        numpy.array: interval index, length L (arbitrary)
        numpy.array: position index, length L (same as interval index)

    Given a set of possibly overlapping intervals, create a mapping of positions that are contained
    within those intervals.

    """

    pos_sort_idx = np.argsort(positions)
    rev_idx = np.arange(len(pos_sort_idx), dtype=int)[pos_sort_idx]

    interval_idx, position_idx = interval_position_overlap(intervals, positions[pos_sort_idx])

    return interval_idx, rev_idx[position_idx]




import bisect
import collections
import pandas as pd
import numpy as np
import scipy.sparse


def create_breakends(breakpoints, data_cols=(), id_col='prediction_id'):
    breakends = breakpoints[[id_col, 'chromosome_1', 'strand_1', 'position_1', 'chromosome_2', 'strand_2', 'position_2']].copy()
    breakends.set_index(id_col, inplace=True)
    breakends.columns = pd.MultiIndex.from_tuples([tuple(c.split('_')) for c in breakends.columns])
    breakends = breakends.stack()
    breakends.index.names = (id_col, 'prediction_side')
    breakends = breakends.reset_index()
    breakends['prediction_side'] = np.where(breakends['prediction_side'] == '1', 0, 1)
    breakends = breakends.merge(breakpoints[[id_col] + list(data_cols)], on=id_col)
    return breakends


class BreakpointDatabase(object):
    def __init__(self, breakpoints, id_col='prediction_id'):
        self.id_col = id_col
        self.breakends = (
            create_breakends(breakpoints, id_col=id_col)
            .sort_values(['chromosome', 'strand', 'position'])
            .set_index(['chromosome', 'strand']))

    def query(self, row, extend=0):
        side_matches = {}

        for side in ('1', '2'):
            chromosome = row['chromosome_' + side]
            strand = row['strand_' + side]
            position = row['position_' + side]

            if (chromosome, strand) not in self.breakends.index:
                return set()

            idx1 = self.breakends.xs((chromosome, strand))['position'].searchsorted(position - extend)
            idx2 = self.breakends.xs((chromosome, strand))['position'].searchsorted(position + extend, side='right')

            side_matches[side] = self.breakends.xs((chromosome, strand)).iloc[idx1:idx2]

        matches = pd.merge(
            side_matches['1'],
            side_matches['2'],
            on=[self.id_col],
            suffixes=('_1', '_2'),
        )

        matches = matches.query('prediction_side_1 != prediction_side_2')

        return set(matches[self.id_col].values)


def match_breakpoints(reference_breakpoints, target_breakpoints, id_col='prediction_id', window_size=500):
    """ Match similar target breakpoints to a set of reference breakpoints  
    """

    reference_db = BreakpointDatabase(reference_breakpoints, id_col=id_col)

    match_data = []

    for idx, row in target_breakpoints.iterrows():
        for reference_id in reference_db.query(row, extend=window_size):
            match_data.append({
                'target_id': row[id_col],
                'reference_id': reference_id,
            })

    match_data = pd.DataFrame(match_data, columns=['target_id', 'reference_id'])

    return match_data


def identify_matched_breakpoints(breakpoints, window_size=500):
    """ Identify common breakpoints.
    
    Args:
        breakpoints (DataFrame): breakpoint data
        
    Returns:
        DataFrame: breakpoints with component_id
        
    The component_id column will have the same value for similar breakpoints
    """

    breakpoints = breakpoints.copy()

    breakpoints['idx'] = range(len(breakpoints.index))

    self_matches = match_breakpoints(
        breakpoints, breakpoints, id_col='idx', window_size=window_size)

    row = self_matches['reference_id'].values
    col = self_matches['target_id'].values
    data = np.ones(len(self_matches.index))
    max_index = self_matches[['reference_id', 'target_id']].max().max()

    assert max_index == breakpoints['idx'].max()

    matrix = scipy.sparse.csr_matrix((data, (row, col)), shape=(max_index + 1, max_index + 1))

    num_components, component_ids = scipy.sparse.csgraph.connected_components(
        matrix, directed=False, return_labels=True)

    components = pd.DataFrame({
        'idx': range(len(component_ids)),
        'component_id': component_ids,
    })

    breakpoints = breakpoints.merge(components, how='outer')

    assert not breakpoints['component_id'].isnull().any()
    assert not breakpoints['idx'].isnull().any()
    
    return breakpoints





import numpy as np
import pandas as pd
import pkg_resources


def read_chromosome_lengths(genome_fasta_index):
    fai = pd.read_csv(genome_fasta_index, sep='\t', header=None, names=['chrom', 'length', 'V3', 'V4', 'V5'])
    fai = fai.set_index('chrom')['length']
    return fai.to_dict()


class RefGenomeInfo(object):
    def __init__(self, version):
        if version == 'hg19':
            self.chromosomes = [str(a) for a in range(1, 23)] + ['X', 'Y']

            genome_fasta_index = pkg_resources.resource_filename('wgs_analysis', 'data/GRCh37-lite.fa.fai')

            self.chromosome_lengths = pd.Series(read_chromosome_lengths(genome_fasta_index)).reindex(self.chromosomes).astype(int)
            self.chromosome_lengths.index.name = 'chr'

            self.chromosome_end = np.cumsum(self.chromosome_lengths)
            self.chromosome_start = self.chromosome_end.shift(1)
            self.chromosome_start[0] = 0
            self.chromosome_start = self.chromosome_start.astype(int)
            self.chromosome_mid = (self.chromosome_start + self.chromosome_end) / 2.

            self.chromosome_info = pd.DataFrame({
                'chromosome_length': self.chromosome_lengths,
                'chromosome_end': self.chromosome_end,
                'chromosome_start': self.chromosome_start,
                'chromosome_mid': self.chromosome_mid,
            }).reset_index()

        elif version == 'grch38':
            self.chromosomes = [f'chr{a}' for a in range(1, 23)] + ['chrX', 'chrY']

            genome_fasta_index = pkg_resources.resource_filename('wgs_analysis', 'data/GRCh38.fa.fai')

            self.chromosome_lengths = pd.Series(read_chromosome_lengths(genome_fasta_index)).reindex(self.chromosomes).astype(int)
            self.chromosome_lengths.index.name = 'chr'

            self.chromosome_end = np.cumsum(self.chromosome_lengths)
            self.chromosome_start = self.chromosome_end.shift(1)
            self.chromosome_start[0] = 0
            self.chromosome_start = self.chromosome_start.astype(int)
            self.chromosome_mid = (self.chromosome_start + self.chromosome_end) / 2.

            self.chromosome_info = pd.DataFrame({
                'chromosome_length': self.chromosome_lengths,
                'chromosome_end': self.chromosome_end,
                'chromosome_start': self.chromosome_start,
                'chromosome_mid': self.chromosome_mid,
            }).reset_index()

info = None


def set_genome_version(version):
    global info
    info = RefGenomeInfo(version)

set_genome_version('hg19')




'''
# Breakpoint primer design

Design primers for breakpoint sequences.  Input to design is a pandas table of breakpoints.  Breakpoint sequences are
created on the fly and single nucleotide changes are masked if available.  Uses the design_primers module to design the
primers and return a table.

'''

import os
import functools
import argparse
import subprocess
import pandas as pd

import design_utils
import design_primers
import blat_server


def get_requirements_filename(experiment_type):
    return os.path.join(os.path.dirname(__file__), '{0}_breakpoint_requirements.tsv'.format(experiment_type))


default_breakpoint_requirements = get_requirements_filename('seqval')


def create_breakpoint_sequences(genome_fasta, breakpoints, variant_vcfs=[]):
    '''
    Generate sequences for a table of breakpoints
    '''
    for idx, row in breakpoints.iterrows():
        seq_id = row['seq_id']
        chroms = row['chrom'].split(',')
        strands = row['strand'].split(',')
        coords = [int(a) for a in row['coord'].split(',')]
        sequences = []
        for side, target_strand in zip((0, 1), ('+', '-')):
            if strands[side] == '+':
                start = coords[side] - 300 + 1
                end = coords[side]
            else:
                start = coords[side]
                end = coords[side] + 300 - 1
            sequence = design_utils.create_sequence(genome_fasta, chroms[side], start, end, variant_vcfs)
            if strands[side] != target_strand:
                sequence = design_utils.reverse_complement(sequence)
            sequences.append(sequence)
        yield seq_id, sequences[0] + '[]' + sequences[1]


def design(genome_filename, breakpoints, variant_vcfs=[], requirements_filename=default_breakpoint_requirements, max_stage=-1, max_primers=-1, primer3_params=None):
    """
    Design primers for an breakpoint validation experiment

    Args:
        genome_filename(str) : genome file in fasta format
        breakpoints(pandas.DataFrame) : table of breakpoint information

    KwArgs:
        variant_vcfs(list) : list of VCF files with germline variants
        requirements_filename(str) : requirements tsv file detailing design requirements by stage
        max_stage(int) : maximum stage before failing, -1 to try all stages
        max_primers(int) : maximum number of primers to design before returning
        primer3_params(dict) : Additional Primer 3 Parameters

    Returns:
        pandas.DataFrame of primer information

    The table of breakpoints should have the following required columns:
     * seq_id : unique identifier for the breakpoint
     * chrom : pair of chromosomes, comma separated
     * strand : pair of strands, comma separated
     * coord : pair of coordinates, comma separated

    """
    breakpoint_sequences = create_breakpoint_sequences(genome_filename, breakpoints, variant_vcfs)

    primer_table = design_primers.design(requirements_filename, breakpoint_sequences, max_stage=max_stage, max_primers=max_primers, primer3_params=primer3_params)

    if len(primer_table.index) == 0:
        return pd.DataFrame(columns=['seq_id']).astype(int)

    def calculate_product_start(row):
        return row['sequence'].index(row['left_primer']) + 1

    def calculate_product_end(row):
        return row['sequence'].index(design_utils.reverse_complement(row['right_primer'])) + len(row['right_primer'])

    def recalculate_sequence(row):
        return row['sequence'][row['product_start'] - 1 : row['product_end']]

    def check_recalculated(row):
        return row['product_sequence'].startswith(row['left_primer']) and \
               row['product_sequence'].endswith(design_utils.reverse_complement(row['right_primer'])) and \
               len(row['product_sequence']) == int(row['product_size'])

    primer_table['product_start'] = primer_table.apply(calculate_product_start, axis=1)
    primer_table['product_end'] = primer_table.apply(calculate_product_end, axis=1)
    primer_table['product_sequence'] = primer_table.apply(recalculate_sequence, axis=1)
    primer_table['product_check'] = primer_table.apply(check_recalculated, axis=1)

    assert primer_table['product_check'].all()

    primer_table = primer_table.drop(['product_sequence', 'product_check'], axis=1)

    primer_table = primer_table.merge(breakpoints, on='seq_id', how='inner')

    return primer_table



if __name__ == '__main__':

    argparser = argparse.ArgumentParser()
    argparser.add_argument('genome', help='genome fasta or 2bit filename')
    argparser.add_argument('breakpoints', help='breakpoints table filename')
    argparser.add_argument('primers', help='primer table filename')
    argparser.add_argument('--requirements', required=False, default=default_breakpoint_requirements, help='design requirements table')
    argparser.add_argument('--varvcfs', nargs='+', required=False, default=[], help='variant vcfs for masking')
    args = argparser.parse_args()

    breakpoints = pd.read_csv(args.breakpoints, sep='\t', converters={'chrom':str})[['chrom', 'coord']].drop_duplicates()

    with blat_server.BlatServer(args.genome):
        primer_table = design(args.genome, breakpoints, args.varvcfs)

    primer_table.to_csv(args.primers, sep='\t', index=False, na_rep='NA')





